/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */

#include "services.h"

/********************************************************************************/


#ifdef CONVERT_29_ANOPE

static void convert_n29();
static void convert_c29();
static void convert_m29();

typedef struct nickinfo29 NickInfo29;

struct nickinfo29 {
	NickInfo29 *next, *prev;
	char nick[NICKMAX];
	char pass[PASSMAX]; char *last_usermask; char *last_realname;
	time_t time_registered; time_t last_seen;
	long accesscount; char **access; long flags;
	time_t id_timestamp;
	unsigned short memomax; unsigned short channelcount;
	char *url; char *email; char *forward; char *hold; char *mark; char *forbid;
	int news; char *regemail; long icq; long auth; long reserved[2];
};


typedef struct {
    short level;
    short is_nick;
    char *name;
    char *sponser;
} ChanAccess29;

typedef struct {
    short is_nick;
    short pad;
    char *name;
    char *reason;
} AutoKick29;

typedef struct chaninfo29 ChanInfo29;
struct chaninfo29 {
    ChanInfo29 *next, *prev;
    char name[64];
    char founder[NICKMAX];              /* Always a reg'd nick */
    char founderpass[PASSMAX];
    char *desc;
    time_t time_registered;
    time_t last_used;
    long accesscount;
    ChanAccess29 *access;                 /* List of authorized users */
    long akickcount;
    AutoKick29 *akick;
    short mlock_on, mlock_off;          /* See channel modes below */
    long mlock_limit;                   /* 0 if no limit */
    char *mlock_key;                    /* NULL if no key */
    char *last_topic;                   /* Last topic on the channel */
    char last_topic_setter[NICKMAX];    /* Who set the last topic */
    time_t last_topic_time;             /* When the last topic was set */
    long flags;                         /* See below */
    short *levels;                      /* Access levels for commands */
    char *url;
    char *email;
    char *welcome;
    char *hold;       /*  }                                         */
    char *mark;       /*  }   --   Identities (what admin did it?)  */
    char *freeze;     /*  }   --                                    */
    char *forbid;     /*  }                                         */
    int topic_allow;                  /* Who's allowed to change topic */
    long auth;
    long reserved[4];                   /* For future expansion -- set to 0 */
};



typedef struct memo29 Memo29;
struct memo29 {
    char sender[NICKMAX]; long number; time_t time; char *text;
    char *chan; short flags; short reserved_s; long reserved[3];
};

typedef struct memolist29 MemoList29;
struct memolist29 {
    MemoList29 *next, *prev; char nick[NICKMAX]; long n_memos; Memo29 *memos; long reserved[4];
};



/********************************************************************************/

E int convert29()
{
	time_t ptime = time(NULL);

	fprintf(stderr,"Sirv 2.9 Database Conversion Session:\n");
	fprintf(stderr,"=====================================\n");
	fprintf(stderr, "Stage 1 - Nick Database");
	convert_n29();
	fprintf(stderr, "\nNick database successfully converted. (%ld second%s)\n\n",
			(time(NULL)-ptime), (time(NULL)-ptime)==1 ? "" : "s");
	fprintf(stderr, "Stage 2 - Chan Database");
	ptime = time(NULL);
	convert_c29();
	fprintf(stderr, "\nChan database successfully converted. (%ld second%s)\n\n",
			(time(NULL)-ptime), (time(NULL)-ptime)==1 ? "" : "s");
	fprintf(stderr, "Stage 3 - Memo Database");
	ptime = time(NULL);
	convert_m29();
	fprintf(stderr, "\nMemo database successfully converted. (%ld second%s)\n\n",
			(time(NULL)-ptime), (time(NULL)-ptime)==1 ? "" : "s");

	fprintf(stderr, "\n*** DO NOT run services with the -convert29 option again.\n");
	fprintf(stderr, "It is highly recommended you run ./configure again and choose NO to "
			"the option of wanting to convert the database and recompile.\n\n");

	return 1;
}

/********************************************************************************/

static u_int16_t read17_16(FILE *f, const char *filename)
{
	int x1, x2;
	u_int16_t i;

	x1 = fgetc(f);
	x2 = fgetc(f);
	if (x1 == EOF || x2 == EOF)
		fatal_error("Error reading database file: %s", filename);

	i = x1 << 8 | x2;

	return i;
}

static u_int32_t read17_32(FILE *f, const char *filename)
{
	int x1, x2, x3, x4;
	u_int32_t i;

	x1 = fgetc(f);	x2 = fgetc(f);	x3 = fgetc(f);	x4 = fgetc(f);
	if (x1 == EOF || x2 == EOF || x3 == EOF || x4 == EOF)
		fatal_error("Error reading database file: %s", filename);

	i = x1 << 24 | x2 << 16 | x3 << 8 | x4;

	return i;
}

static char *read17(FILE *f, const char *filename)
{
	int x1, x2;
	char *s;
	int len;

	x1 = fgetc(f);
	x2 = fgetc(f);
	if (x1 == EOF || x2 == EOF)
		fatal_error("Error reading database file: %s", filename);

	len = x1 << 8 | x2;

	if (!len)
		return NULL;
	s = smalloc(len);

	if (len != fread(s, 1, len, f))
		fatal_error("Database conversion failed. Read Error on file =  %s",filename);

	return s;

}

static char *read29(FILE *f, const char *filename)
{
	char *s;
	int len;

	len = fgetc(f) * 256 + fgetc(f);
	s = smalloc(len);
	if (len != fread(s, 1, len, f))
		fatal_error("Database conversion failed. Read Error on file =  %s",filename);

	return s;


}

/********************************************************************************/

/* type 1 = sirv2.9, type 2 = anope 1.7 nick version, type 3 = anope 1.7 chan version */


static FILE *check_version(const char *filename, int type)
{
	FILE *f;
	int ver;

	if (!(f = fopen(filename, "r")))
		fatal_error("Ddatabase conversion failed. Can't locate file = %s", filename);

	ver = fgetc(f)<<24 | fgetc(f)<<16 | fgetc(f)<<8 | fgetc(f);

	if (ferror(f))
		fatal_error("Database conversion failed. Error reading file = %s", filename);
	else if (ferror(f))
		fatal_error("Database conversion failed. Unexpected end of file reached for = %s", filename);
	else if (type == 1 && ver != 8)
		fatal_error("Database conversion failed. Invalid version number (%d) for file =  %s", ver, filename);
	else if (type == 2 && ver != 13)
		fatal_error("Database conversion failed. Invalid version number (%d) for file =  %s", ver, filename);
	else if (type == 3 && ver != 16)
		fatal_error("Database conversion failed. Invalid version number (%d) for file =  %s", ver, filename);


	return f;
}


/********************************************************************************/

/* Convert sirv 2.9 nick database into sirv 3's SQL format. */

static void convert_n29()
{
	FILE *f;
	char filename[strlen(DATDEST) + 10];
	int i;
	NickInfo29 *ni;
	unsigned long id=1;

	sprintf(filename, "%s/nick.db", DATDEST);


	f = check_version(filename, 1);


	for (i=33; i < 256; i++) {
		fprintf(stderr, ".");
		while (fgetc(f) == 1) {
			long flags=0;
			ni = smalloc(sizeof(NickInfo29));


			if (fread(ni, sizeof(NickInfo29), 1, f) != 1)
				fatal_error("Nick database conversion failed. Error reading file = %s", filename);



			send_sql("INSERT INTO nick VALUES "
				"(%d,'%s','%s',NULL,NULL,NULL,'temp@host.com','%d','%d',0,0,NULL,NULL,-1,0,0)",
				id++,ni->nick, ni->pass, ni->time_registered, ni->last_seen);



			if (ni->url)
				send_sql("UPDATE nick SET url = '%s' WHERE id = %d",
					read29(f, filename), id-1);

			if (ni->email)
				send_sql("UPDATE nick SET email = '%s' WHERE id = %d",
					read29(f, filename), id-1);

			if (ni->forward)
				send_sql("UPDATE nick SET forward = '%s' WHERE id = %d",
					read29(f, filename), id-1);

			if (ni->hold) {
				send_sql("INSERT INTO flags VALUES ('%s',%d,0,0,0)",
					ni->nick, time(NULL));
				send_sql("INSERT INTO reason VALUES "
					"('%s','HELD','No Reason Specified.','%s')", 
					ni->nick, read29(f, filename));
				flags |= NS_HELD;
			}

			if (ni->mark) {
				if (ni->hold)
					send_sql("UPDATE flags SET marked = %d WHERE name = %s",
						time(NULL), ni->nick);
				else
					send_sql("INSERT INTO flags VALUES ('%s',0,%d,0,0)",
						ni->nick, time(NULL));

				send_sql("INSERT INTO reason VALUES "
					"('%s','MARKED','No Reason Specified.','%s')", 
					ni->nick, read29(f, filename));
			}

			if (ni->forbid)
				ni->forbid = read29(f, filename);


			if (ni->regemail)
				ni->regemail = read29(f, filename);

			send_sql("UPDATE nick SET usermask = '%s' WHERE id = %d",
				read29(f, filename), id-1);
			ni->last_realname = read29(f, filename);


			if (ni->accesscount) {
				int x;

				for (x=0; x < ni->accesscount; x++)
					send_sql("INSERT INTO naccess VALUES (%d, '%s')",
						id-1, read29(f, filename));

			}

			if (ni->flags & 0x00000001)
				flags |= NS_ENFORCE;
			if (ni->flags & 0x00000010)
				flags |= MS_MEMO_LOGON;
			if (ni->flags & 0x00000020)
				flags |= MS_MEMO_NEW;
			if (ni->flags & 0x00000002)
				flags |= NS_SECURE;
			if (ni->flags & 0x00000100)
				flags |= NS_PRIVATE;
			if (ni->flags & 0x00001000)
				flags |= MS_NOMEMO;
			if (ni->flags & 0x00000800)
				flags |= NS_NEVEROP;
			
			
			send_sql("UPDATE nick SET flags = %d WHERE id = %d",
				flags, id-1);


			if (ni->forbid) {
				send_sql("DELETE FROM nick WHERE name = '%s'", ni->nick);
				send_sql("INSERT into nick VALUES "
					"(%d,'%s','*','*','*','*','*',0,0,-1,0,NULL,NULL,0,NULL,0)",
						id-1, ni->nick);
				send_sql("INSERT INTO flags VALUES ('%s',0,0,'%d',0)",
					ni->nick, time(NULL));
				send_sql("INSERT INTO reason VALUES "
					"('%s','FORBIDDEN','Reason not specified.','%s')", 
					ni->nick, ni->forbid);


			}


			free(ni);

		}
	}

	fclose(f);
/* Some entries will actually be NULL, which will break sirv3's forward code. This step is vital. */
	send_sql("UPDATE nick SET forward='0' WHERE forward IS NULL");
}

/********************************************************************************/

static void convert_c29()
{
	FILE *f;
	char filename[strlen(DATDEST) + 10];
	int i;
	ChanInfo29 *ci;
	unsigned long id=1;

	sprintf(filename, "%s/chan.db", DATDEST);

	f = check_version(filename, 1);

	for (i=33; i < 256; i++) {
		fprintf(stderr, ".");
		while (fgetc(f) == 1) {
			long flags=0, modeson=0, modesoff=0, fid=0;
			ci = smalloc(sizeof(ChanInfo29));
			if (fread(ci, sizeof(ChanInfo29), 1, f) != 1)
				fatal_error("Chan database conversion failed. Error reading file = %s", filename);


			if (sql_select("id FROM nick WHERE name = '%s'", ci->founder)) {
				row = fetch(res);
				fid = atol(row[0]);
			}

			send_sql("INSERT INTO chan VALUES "
				"(%d,'%s','%s','%s',%d,'temp',NULL,NULL,%d,%d,NULL,NULL,"
				"0,0,0,0,NULL,NULL,NULL,0,5)", id++,
				ci->name, ci->founderpass, ci->founder, fid, ci->time_registered, ci->last_used);


			send_sql("UPDATE chan SET descr = '%s' WHERE id = %d",
				read29(f, filename), id-1);


			if (ci->url)
				send_sql("UPDATE chan SET url = '%s' WHERE id = %d",
					read29(f, filename), id-1);
			if (ci->email)
				read29(f, filename);

			if (ci->mlock_key)
				send_sql("UPDATE chan SET chankey = '%s' WHERE id = %d",
				read29(f, filename), id-1);

			if (ci->last_topic)
				send_sql("UPDATE chan SET topic = '%s', tsetter = '%s', "
					"ttime = %d WHERE id = %d",
					read29(f, filename), s_ChanServ, ci->last_topic_time, id-1);



			if (ci->welcome)
				send_sql("UPDATE chan SET welcome = '%s' WHERE id = %d",
					read29(f, filename), id-1);
			if (ci->hold) {
				send_sql("INSERT INTO cflags VALUES ('%s',%d,0,0,0,0)",
					ci->name, time(NULL));
				send_sql("INSERT INTO reason VALUES "
					"('%s','HELD','No Reason Specified.','%s')", 
					ci->name, read29(f, filename));
				flags |= NS_HELD;
			}

			if (ci->mark) {
				if (ci->hold)
					send_sql("UPDATE cflags SET marked = %d WHERE name = '%s'",
						time(NULL), ci->name);
				else
					send_sql("INSERT INTO cflags VALUES ('%s',0,%d,0,0,0)",
						ci->name, time(NULL));

				send_sql("INSERT INTO reason VALUES "
					"('%s','MARKED','No Reason Specified.','%s')", 
					ci->name, read29(f, filename));

			}

			if (ci->freeze) {
				if (ci->hold || ci->mark)
					send_sql("UPDATE cflags SET frozen = %d WHERE name = '%s'",
						time(NULL), ci->name);
				else
					send_sql("INSERT INTO cflags VALUES ('%s',0,0,0,%d,0)",
						ci->name, time(NULL));

				send_sql("INSERT INTO reason VALUES "
					"('%s','FROZEN','No Reason Specified.','%s')", 
					ci->name, read29(f, filename));
			}

			if (ci->forbid) {
				if (ci->hold || ci->mark || ci->freeze)
					send_sql("UPDATE cflags SET closed = %d WHERE name = '%s'",
						time(NULL), ci->name);
				else
					send_sql("INSERT INTO cflags VALUES ('%s',0,0,%d,0,0)",
						ci->name, time(NULL));

				send_sql("INSERT INTO reason VALUES "
					"('%s','CLOSED','No Reason Specified.','%s')", 
					ci->name, read29(f, filename));
			}

			if (ci->accesscount) {
				ChanAccess29 *access = smalloc(sizeof(ChanAccess29) * ci->accesscount);
				int x;

				if (ci->accesscount != fread(access, sizeof(ChanAccess29), ci->accesscount, f))
					fatal_error("Chan database conversion failed. "
						 "Error reading file = %s (Access list)", filename);


				for (ci->access=access, x=0; x < ci->accesscount; ++x, ++access) {

					access->name = read29(f,filename);
					access->sponser = read29(f,filename);

					send_sql("INSERT INTO caccess VALUES "
						"(%d,'%s',%d,'%s',%d,NULL)", id-1, access->name, 
						access->level, access->sponser, time(NULL));
				}

			}
			if (ci->akickcount) {
				AutoKick29 *akick = smalloc(sizeof(AutoKick29) * ci->akickcount);
				int x;


				if (ci->akickcount != fread(akick, sizeof(AutoKick29), ci->akickcount, f))
					fatal_error("Chan database conversion failed. "
						 "Error reading file = %s (AKick list)", filename);


				for (ci->akick = akick, x=0; x < ci->akickcount; ++x, ++akick) {
					akick->name = read29(f,filename);

					send_sql("INSERT INTO caccess VALUES "
						"(%d,'%s',-1,'%s',%d,'%s')", 
						id-1, akick->name, s_ChanServ, time(NULL), 
						akick->reason ? read29(f,filename) : "Not Specified");


				}
			}

			
			if (ci->levels) {
				int entries = fgetc(f)<<8 | fgetc(f);

				if (entries < 0)
					fatal_error("Chan database conversion failed. "
						 "Error reading file = %s (levels)", filename);

				fseek(f, sizeof(short) * entries, SEEK_CUR);
			}

			if (ci->flags & 0x00000001)
				flags |= CS_KEEPTOPIC;
			if (ci->flags & 0x00000002)
				flags |= CS_OPGUARD;
			if (ci->flags & 0x00000004)
				flags |= CS_PRIVATE;
			if (ci->flags & 0x00000010)
				flags |= CS_RESTRICTED;
			if (ci->flags & 0x00000020)
				flags |= CS_LEAVEOPS;
			if (ci->flags & 0x00000040)
				flags |= CS_IDENT;
			if (ci->flags & 0x01000000)
				flags |= CS_UNSECURE;


			modeson |= CMODE_R;
			if (ci->mlock_on & 0x00000001)
				modeson |= CMODE_I;
			if (ci->mlock_on & 0x00000002)
				modeson |= CMODE_M;
			if (ci->mlock_on & 0x00000004)
				modeson |= CMODE_N;
			if (ci->mlock_on & 0x00000008)
				modeson |= CMODE_P;
			if (ci->mlock_on & 0x00000010)
				modeson |= CMODE_S;
			if (ci->mlock_on & 0x00000020)
				modeson |= CMODE_T;
			if (ci->mlock_on & 0x00000040)
				modeson |= CMODE_K;
			if (ci->mlock_on & 0x00000080)
				modeson |= CMODE_L;
			if (ci->mlock_on & 0x00000200)
				modeson |= CMODE_J;
			if (ci->mlock_on & 0x00000400)
				modeson |= CMODE_r;
			if (ci->mlock_on & 0x00000800)
				modeson |= CMODE_C;
			if (ci->mlock_on & 0x00001000)
				modeson |= CMODE_o;
			if (ci->mlock_on & 0x00002000)
				modeson |= CMODE_m;



			if (ci->mlock_off & 0x00000001)
				modesoff |= CMODE_I;
			if (ci->mlock_off & 0x00000002)
				modesoff |= CMODE_M;
			if (ci->mlock_off & 0x00000004)
				modesoff |= CMODE_N;
			if (ci->mlock_off & 0x00000008)
				modesoff |= CMODE_P;
			if (ci->mlock_off & 0x00000010)
				modesoff |= CMODE_S;
			if (ci->mlock_off & 0x00000020)
				modesoff |= CMODE_T;
			if (ci->mlock_off & 0x00000040)
				modesoff |= CMODE_K;
			if (ci->mlock_off & 0x00000080)
				modesoff |= CMODE_L;
			if (ci->mlock_off & 0x00000200)
				modesoff |= CMODE_J;
			if (ci->mlock_off & 0x00000400)
				modesoff |= CMODE_r;
			if (ci->mlock_off & 0x00000800)
				modesoff |= CMODE_C;
			if (ci->mlock_off & 0x00001000)
				modesoff |= CMODE_o;
			if (ci->mlock_off & 0x00002000)
				modesoff |= CMODE_m;


			send_sql("UPDATE chan SET flags = %d, modeson = %d, modesoff = %d WHERE id = %d",
				flags, modeson, modesoff, id-1);




			free(ci);
		}
	}

/* Clean up any empty access list entries that could pose to be a problem later on */
	send_sql("DELETE FROM caccess WHERE entry = ''");

}

/********************************************************************************/

static void convert_m29()
{
	FILE *f;
	MemoList29 *ml;
	Memo29 *m;
	char filename[strlen(DATDEST) + 10];
	int i,x;

	sprintf(filename, "%s/memo.db", DATDEST);


	f = check_version(filename, 1);


	for (i=33; i < 256; i++) {
		fprintf(stderr, ".");
		while (fgetc(f) == 1) {
			ml = smalloc(sizeof(MemoList29));
			if (fread(ml, sizeof(MemoList29), 1, f) != 1)
				fatal_error("Memo database conversion failed. Error reading file = %s", filename);


			ml->memos = m = smalloc(sizeof(Memo29) * ml->n_memos);
			fread(m, sizeof(Memo29), ml->n_memos, f);
			for (x=0; x < ml->n_memos; ++x, ++m) {
				int new=0, del=0;

				m->text = read29(f, filename);

				if (m->flags & 0x0001)
					new=1;
				if (m->flags & 0x0002)
					del=1;

				if (m->chan) {
					m->chan = read29(f, filename);

					send_sql("INSERT INTO memos VALUES "
						"('%s','%s','%s','%s',%d,%d,%d,0)", ml->nick, m->sender,
						m->chan, m->text, m->time, new, del);


				} else
					send_sql("INSERT INTO memos VALUES ('%s','%s',NULL,'%s',%d,0,0,0)",
						ml->nick, m->sender, m->text, m->time);
			}
		}
	}

}


/********************************************************************************/
/********************************************************************************/

int convert_anope17();


static void convert_n17();
static void convert_c17();


/********************************************************************************/

int convert17() {

	time_t ptime = time(NULL);

	fprintf(stderr,"Anope 1.7 Database Conversion Session:\n");
	fprintf(stderr,"=====================================\n");
	fprintf(stderr, "Stage 1 - Nick & Memo Database");

	convert_n17();
	fprintf(stderr, "\nNick database successfully converted. (%ld second%s)\n\n",
			(time(NULL)-ptime), (time(NULL)-ptime)==1 ? "" : "s");


	ptime = time(NULL);
	fprintf(stderr, "Stage 2 - Chan Database");
	convert_c17();
	fprintf(stderr, "\nChan database successfully converted. (%ld second%s)\n\n",
			(time(NULL)-ptime), (time(NULL)-ptime)==1 ? "" : "s");



	fprintf(stderr, "\n*** DO NOT run services with the -convert17 option again.\n");
	fprintf(stderr, "It is highly recommended you run ./configure again and choose NO to "
			"the option of wanting to convert the database and recompile.\n\n");

	return 1;

}

/********************************************************************************/

static void convert_n17()
{
	FILE *f;
	char filename[strlen(DATDEST) + 10];
	char *name, *pass, *s, *mask;
	int i, y=0, z=0, num_rows=0;
	long flags=0, xflags=0, icq=0;
	unsigned long id=1;
#ifdef USE_ENCRYPTION
	char encrypt_pass[16];
#endif

	sprintf(filename, "%s/nick.db", DATDEST);

	f = check_version(filename, 2);

	for (i=0; i < 1024; i++) {
		fprintf(stderr, ".");
		while (fgetc(f) == 1) {
			flags=0;
			name = read17(f,filename);
			pass = read17(f,filename);

#ifdef USE_ENCRYPTION
			if (pass) {
				memcpy(encrypt_pass, pass, 16);
				encrypt_pass[16]=0;
				pass = encrypt_pass;
			}
#endif

			send_sql("INSERT INTO nick VALUES "
				"(%d,'%s','%s',NULL,NULL,NULL,'temp@host.com',0,0,0,0,NULL,NULL,-1,0,0)",
				id++,name, pass ? pass : "!");


			if ((s = read17(f,filename))) {
				send_sql("UPDATE nick SET email = '%s' WHERE name = '%s'",
					s, name);
				free(s);
			}

			read17(f, filename);  /* greet */


			if ((icq = read17_32(f, filename)))
				send_sql("UPDATE nick SET icq = %d WHERE name = '%s'",
					icq, name);


			if ((s = read17(f,filename))) {
				send_sql("UPDATE nick SET url = '%s' WHERE name = '%s'",
					s, name);
				free(s);
			}


			xflags = read17_32(f, filename); /* flags */


			if (xflags & 0x00000001)
				flags |= NS_ENFORCE;
			if (xflags & 0x00000010)
				flags |= MS_MEMO_LOGON;
			if (xflags & 0x00000020)
				flags |= MS_MEMO_NEW;
			if (xflags & 0x00000002)
				flags |= NS_SECURE;
			if (xflags & 0x00000040)
				flags |= NS_PRIVATE;
			if (xflags & 0x00000080)
				flags |= NS_HIDE_EMAIL;
			if (xflags & 0x00004000)
				flags |= NS_ENCRYPTED;
			if (xflags & 0x00010000)
				flags |= MS_EMAILMEMOS;


			send_sql("UPDATE nick SET flags = %d WHERE name = '%s'",
				flags, name);


			read17_16(f, filename); /* language */

			z=read17_16(f, filename); /* access count */


			if (z) {
				int j;
				for (j=0; j < z; j++) {
					s = read17(f, filename);
					if (s) {
						send_sql("INSERT INTO naccess VALUES (%d, '%s')",
							id-1, s);
						free(s);
					}
				}
			}

			z=read17_16(f, filename); /* memo count */
			read17_16(f, filename); /* memo max */

			if (z) {
				int j;
				char buf[32];
				for (j=0; j < z; j++) {
					time_t mtime;
					int flagnew=0, v=0;

					read17_32(f, filename); /* index */
					v=read17_16(f, filename); /* flags */
					mtime=read17_32(f, filename); /* time */

					if (v & 0x001)
						flagnew=1;


					fread(buf,1,sizeof(buf),f); /* sender */
					s=read17(f, filename); /* text */

					if (s) {
						send_sql("INSERT INTO memos VALUES "
							"('%s','%s',NULL,'%s',%d,%d,0,0)",
							name, buf, s, mtime, flagnew);
						free(s);
					}
				}
			}

			read17_16(f, filename); /* channel count */
			read17_16(f, filename); /* channel max */

			free(name); 

		}
	}

	for (i=0; i < 1024; i++) {
		while ((y = fgetc(f) == 1)) {
			long rtime=0, ltime=0, sflags=0;
			name = read17(f, filename);
			mask = read17(f, filename);


			if ((s = read17(f, filename))) /* real name */
				free(s);

			if ((s = read17(f, filename))) /* last quit */
				free(s);


			rtime = read17_32(f, filename); /* reg time */
			ltime = read17_32(f, filename); /* last seen */


			sflags = (read17_16(f, filename) & 0x0002); /* status */

			s = read17(f, filename); /* core */


			if (sflags) {
				send_sql("DELETE FROM nick WHERE name = '%s'", name);
				send_sql("INSERT into nick VALUES "
					"(%d,'%s','*','*','*','*','*',0,0,-1,0,NULL,NULL,0,0,0)",
						id++, name);
				send_sql("INSERT INTO flags VALUES ('%s',0,0,'%d',0)",
					name, time(NULL));
				send_sql("INSERT INTO reason VALUES "
					"('%s','FORBIDDEN','Reason not specified.','%s')", 
					name, s_NickServ);

				if (sstrcmp(s, name)) {
					send_sql("INSERT into nick VALUES "
						"(%d,'%s','*','*','*','*','*',0,0,-1,0,NULL,NULL,0,0,0)",
							id++, s);
					send_sql("INSERT INTO flags VALUES ('%s',0,0,'%d',0)",
						s, time(NULL));
					send_sql("INSERT INTO reason VALUES "
						"('%s','FORBIDDEN','Reason not specified.','%s')", 
						s, s_NickServ);
				}


			} else if (sstrcmp(s, name)) {   /* grouped nick */

				if (sql_select("pass,email,icq,url,flags FROM nick WHERE name = '%s'", s)) {
					row = fetch(res);

					send_sql("INSERT INTO nick VALUES "
					"(%d,'%s','%s',NULL,NULL,NULL,'%s',%d,%d,0,0,NULL,NULL,-1,0,0)",
						id++,name, row[0], mask, rtime, ltime);

					if (row[1])
						send_sql("UPDATE nick SET email = '%s' WHERE name = '%s'",
							row[1], name);
					if (row[2])
						send_sql("UPDATE nick SET icq = '%s' WHERE name = '%s'",
							row[2], name);
					if (row[3])
						send_sql("UPDATE nick SET url = '%s' WHERE name = '%s'",
							row[3], name);
					if (row[4])
						send_sql("UPDATE nick SET flags = %d WHERE name = '%s'",
							atol(row[4]));

				}

				if ((num_rows=sql_select("mask FROM naccess WHERE name = '%s'", s))) {
					int j;
					for (j=0; j < num_rows; j++) {
						row = fetch(res);
						send_sql("INSERT INTO naccess VALUES (%d, '%s')",
							id-1, row[j]);
					}

				}
			} else if (mask) {    /* A valid core nick (not forbidden) */

				send_sql("UPDATE nick SET usermask = '%s', "
					"time = %d, lasttime = %d WHERE name = '%s'",
					mask, rtime, ltime, name);

			}


			free(name); free(s);
			if (mask)
				free(mask);

		}
	}

}



static void convert_c17()
{
	FILE *f;
	char filename[strlen(DATDEST) + 10];
	char *s;
	int i, x, z;
	unsigned long id=1, zz;
#ifdef USE_ENCRYPTION
	char encrypt_pass[16];
#endif

	sprintf(filename, "%s/chan.db", DATDEST);

	f = check_version(filename, 3);

	for (i=0; i < 256; i++) {
		fprintf(stderr, ".");

		while (fgetc(f) != 0) {
		char cname[64], nick[32], pass[32];
		long xflags=0, flags=0, modeson=0, modesoff=0, fid=0;

		*cname=0; *nick=0; *pass=0;

		fread(cname,1,sizeof(cname),f);

		s = read17(f, filename);


		if (s) {
			if (sql_select("id FROM nick WHERE name = '%s'", s)) {
				row = fetch(res);
				fid = atol(row[0]);
			}
		}

		send_sql("INSERT INTO chan VALUES "
			"(%d,'%s','*','%s',%d,'*',NULL,NULL,0,0,NULL,NULL,"
			"0,0,0,0,NULL,NULL,NULL,0,5)",
			id++,cname, s ? s : "*", fid);


		if ((s = read17(f,filename))) {
			send_sql("UPDATE chan SET successor = '%s' WHERE id = %d",
				s, id-1);
		}


		if (fread(pass,1,sizeof(pass),f) ) {
#ifdef USE_ENCRYPTION
			char *passwd;
			memcpy(encrypt_pass, pass, 16);
			encrypt_pass[16]=0;
			passwd = encrypt_pass;
			send_sql("UPDATE chan SET pass = '%s' WHERE id = %d",
				passwd, id-1);
#else
			send_sql("UPDATE chan SET pass = '%s' WHERE id = %d",
				pass, id-1);
#endif
		}

		s = read17(f,filename);
		send_sql("UPDATE chan SET descr = '%s' WHERE id = %d",
			s ? s : "*", id-1);


		if ((s = read17(f,filename))) {
			send_sql("UPDATE chan SET url = '%s' WHERE id = %d",
				s, id-1);
		}

		read17(f,filename); /* email */


		send_sql("UPDATE chan SET time = %d WHERE id = %d",
				read17_32(f,filename), id-1);

		send_sql("UPDATE chan SET lasttime = %d WHERE id = %d",
				read17_32(f,filename), id-1);



		if ((s = read17(f,filename))) {
			if (sstrcmp(s, ""))
				send_sql("UPDATE chan SET topic = '%s' WHERE id = %d",
					s, id-1);
		}

		if (fread(nick,1,sizeof(nick),f) ) {
			if (sstrcmp(nick, ""))
				send_sql("UPDATE chan SET tsetter = '%s' WHERE id = %d",
					nick, id-1);
		}

		if ((zz = read17_32(f, filename)))
			send_sql("UPDATE chan SET ttime = %d WHERE id = %d",
				zz, id-1);


		xflags = read17_32(f, filename);

		if (xflags & 0x00000001)
			flags |= CS_KEEPTOPIC;
		if (xflags & 0x00000002)
			flags |= CS_OPGUARD;
		if (xflags & 0x00000004)
			flags |= CS_PRIVATE;
		if (xflags & 0x00000010)
			flags |= CS_RESTRICTED;
		if (xflags & 0x00000040)
			flags |= CS_IDENT;
		if (xflags & 0x00000100)
			flags |= CS_ENCRYPTED;

		send_sql("UPDATE chan SET flags = %d WHERE id = %d",
			flags, id-1);


		read17(f, filename);  /* forbid by */
		read17(f, filename); /* forbid reason */

		read17_16(f, filename); /* ban type */
		z = read17_16(f, filename); /* levels */

		for (x=0; x < z; x++) {
			read17_16(f, filename);
		}

		z = read17_16(f, filename); /* access list count */

		if (z) {
			int j;

			for (j=0; j < z; j++) {
				if (read17_16(f, filename)) {    /* in use */
					int level=0;

					level = read17_16(f, filename);
					s = read17(f, filename);
					if (s && atoi(s) != -2) {
						send_sql("INSERT INTO caccess VALUES "
						"(%d,'%s',%d,'%s',%d,NULL)",
						id-1, s, level, s_ChanServ, time(NULL));
					}
					read17_32(f, filename); /* last seen */
				}
			}
		}  /* if (z) access listing */

		z = read17_16(f, filename);  /* akick acount */

		if (z) {
			int j, flags;
			for (j=0; j < z; j++) {
				int level = -1;
				char *entry;
				flags=read17_16(f, filename);
				if (flags & 0x0001) {
					entry = read17(f, filename);
					if (flags & 0x0002) { /* nick entry */
						level=-2;
					}

					send_sql("INSERT INTO caccess VALUES "
						"(%d,'%s',%d,'%s',%d,NULL)",
						id-1, entry, level, s_ChanServ, time(NULL));

					if ((s = read17(f, filename))) {
						send_sql("UPDATE caccess SET reason='%s' "
							"WHERE chan = %d AND entry = '%s' AND level = -1",
							s, id-1, entry);
					}

					if ((s = read17(f, filename))) {
						send_sql("UPDATE caccess SET instater='%s' "
							"WHERE chan = %d AND entry = '%s' AND level = -1",
							s, id-1, entry);
					}

					if ((zz = read17_32(f, filename))) {
						send_sql("UPDATE caccess SET time='%d' "
							"WHERE chan = %d AND entry = '%s' AND level = -1",
							(flags & 0x0001) ? zz : time(NULL), id-1, entry);
					}

				}

			}
		}  /* if (z)  akick list  */

		zz = read17_32(f, filename);  /* mlocks on */

		modeson |= CMODE_R;
		if (zz & 0x00000001)
			modeson |= CMODE_I;
		if (zz & 0x00000002)
			modeson |= CMODE_M;
		if (zz & 0x00000004)
			modeson |= CMODE_N;
		if (zz & 0x00000008)
			modeson |= CMODE_P;
		if (zz & 0x00000010)
			modeson |= CMODE_S;
		if (zz & 0x00000020)
			modeson |= CMODE_T;
		if (zz & 0x00000040)
			modeson |= CMODE_K;
		if (zz & 0x00000080)
			modeson |= CMODE_L;
		if (zz & 0x00000400)
			modeson |= CMODE_C;
		if (zz & 0x00000800)
			modeson |= CMODE_m;
		if (zz & 0x00008000)
			modeson |= CMODE_o;


		zz = read17_32(f, filename);  /* mlocks off */



		if (zz & 0x00000001)
			modesoff |= CMODE_I;
		if (zz & 0x00000002)
			modesoff |= CMODE_M;
		if (zz & 0x00000004)
			modesoff |= CMODE_N;
		if (zz & 0x00000008)
			modesoff |= CMODE_P;
		if (zz & 0x00000010)
			modesoff |= CMODE_S;
		if (zz & 0x00000020)
			modesoff |= CMODE_T;
		if (zz & 0x00000040)
			modesoff |= CMODE_K;
		if (zz & 0x00000080)
			modesoff |= CMODE_L;
		if (zz & 0x00000400)
			modesoff |= CMODE_C;
		if (zz & 0x00000800)
			modesoff |= CMODE_m;
		if (zz & 0x00008000)
			modesoff |= CMODE_o;

		zz=read17_32(f, filename);  /* mlock limit */
		s = read17(f, filename);  /* mlock key */

		if (zz < 1 && (modeson & CMODE_L))
			modeson &= ~CMODE_L;

		if (!s && (modeson & CMODE_K))
			modeson &= ~CMODE_K;


		send_sql("UPDATE chan SET modeson = %d, modesoff = %d WHERE id = %d",
			modeson, modesoff, id-1);

		if (zz > 0 && (modeson & CMODE_L))
			send_sql("UPDATE chan SET chanlimit = %d WHERE id = %d", zz, id-1);

		if (s && (modeson & CMODE_K))
			send_sql("UPDATE chan SET chankey = '%s' WHERE id = %d", s, id);


		read17(f, filename);  /* fmode */

		read17(f, filename);  /* lmode */


		z=read17_16(f, filename);  /* memo count */
		read17_16(f, filename);  /* memo max */


		if (z) {
			int j;

			for (j=0; j < z; j++) {
				char buf[32];
				read17_32(f, filename); /* number */
				read17_16(f, filename); /* flags */
				read17_32(f, filename); /* time */
				fread(buf,1,sizeof(buf),f); /* sender */
				read17(f, filename); /* text */
			}
		}



		read17(f, filename);


		read17(f, filename); /* botserv */

		read17_32(f, filename); /* bot flags */
		z=read17_16(f, filename);

		for (x=0; x < z; x++) {
			read17_16(f, filename);
		}

		read17_16(f, filename);
		read17_16(f, filename);
		read17_16(f, filename);
		read17_16(f, filename);
		read17_16(f, filename);
		z=read17_16(f, filename);

		if (z) {
			int j;
			for (j=0; j < z; j++) {
				zz=read17_16(f, filename);
				if (zz) {
					read17(f, filename);
					read17_16(f, filename);
				}
			}
		}

		}
	}
	send_sql("DELETE FROM caccess WHERE entry = ' '");
	send_sql("DELETE FROM chan WHERE founder = '*'");
}

/********************************************************************************/
/********************************************************************************/


#endif /* CONVERT_29_ANOPE */

