/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */


#include "services.h"
#include "log.h"

CloseChans *closelist = NULL;

ChannelNode *channelnodes[256];
ChannelNode *chanlist, *end;


typedef struct channode ChanNode;
struct channode {
	ChanNode *left, *right;
	ChanInfo *ci;
	unsigned long value;
};

ChanNode *channodes[256], *ci_list=NULL, *officials=NULL;


void chan_adduser(const char *chan, User *client, time_t stamp);


unsigned int cchans=0, maxchans=0, monitored=0;

/***********************************************************************************************/
/***********************************************************************************************/


static void resynch_channel(Channel *c)
{
	struct ChanUsers *cu = c->users, *cu2;


	if (!cu && debug)
		log("Attempted to resynch channel %s with no active clients!", c->name);

	if (debug)
		log("Resynching channel %s", c->name);


	while (cu) {
		cu2 = cu->next;
		chan_deluser(cu->user, c);
		cu = cu2;
	}

	send_cmd(NULL, "RESYNCH %s", c->name);

}


/***********************************************************************************************/

void m_sjoin(char *source, int ac, char **av)
{
	Channel *c;
	char *users, *chan, *cptr;
	char *mv[5];
	User *u;
	int valid_user=0;       /* if there is at least 1 valid user, process channel modes */
	int mvfree=0;
	time_t stamp = atol(av[0]);


#ifdef BAHAMUT
	if (!strchr(source, '.')) {
		users = source;
		chan = av[ac-1];
	} else {
#else
	{
#endif
		users = av[ac-1];
		chan = av[1];
		mv[0] = sstrdup(chan);
		mvfree=3;

		if (ac > 3)
			mv[2] = sstrdup(av[2]);
#ifdef UNREAL
		else
			mv[2] = sstrdup("+");
#endif

		if (ac > 4 && av[3])
			mv[mvfree++] = sstrdup(av[3]);
		if (ac > 5 && av[4])
			mv[mvfree++] = sstrdup(av[4]);
		if (ac > 6 && av[5])
			mv[mvfree++] = sstrdup(av[5]);
	}

	if ((c = findchan(chan))) {  /* check for ts discrepancy and resynch if required. */
		if (c->stamp > stamp) {
			resynch_channel(c);
			return;
		}
	} else if (!c && av[ac-2] == "0")  /* waiting for resynch, ignore all other burst sjoins for this chan */
			return;



	while (*users) {
		int isop=0, isvoiced=0;
#ifdef HAS_HALFOP
		int ishalf=0;
#endif
#ifdef UNREAL
		int isban=0;
#endif



		cptr = strchr(users, ' ');

		if (cptr)
			*cptr++ = 0;

		if (*users == '@') {
			*users++;
			isop=1;
		}
		if (*users == '+') {
			*users++;
			isvoiced=1;
		}
#ifdef HAS_HALFOP
		if (*users == '%') {
			*users++;
			ishalf=1;
		}
#endif
#ifdef UNREAL
		if (*users == '*') {
			*users++;
		}
		if (*users == '~') {
			*users++;
		}
		if (*users == '&') {
			*users++;
			chan_addban(chan, users, 0);
			isban=1;
		}
		if (*users == '"') {
			*users++;
			isban=1;
		}
		if (isban) {
			if (cptr)
				users=cptr;
			else
				*users=0;
			continue;
		}
#endif




		if (!(u = finduser(users))) {
			log("SJOIN: Nonexistent user '%s' for channel %s", users, chan);


		} else if (cburst && check_should_kick(u, chan)) {
			;

		} else {
			valid_user++;
			check_welcome(u->nick, chan);
			chan_adduser(chan, u, stamp);
			if (isvoiced)
				chan_addop(chan, u->nick, 1);


			if (cburst && isop) {
				if (check_valid_op(u, chan)) {
					chan_addop(chan, u->nick, 0);

#ifdef UNREAL
				check_special_op(u, chan);
#endif
				}
			} else if (cburst && (!isop || !isvoiced)) {
				int acc=check_should_op(u, chan);

				if (!isvoiced && (acc==3))
					chan_addop(chan, u->nick, 1);
				else if (!isop && acc==1)
					chan_addop(chan, u->nick, 0);
#ifdef HAS_HALFOP
				else if (!ishalf && acc==4)
					chan_addop(chan, u->nick, 2);
#endif

			} else if (!cburst && isop)
				chan_addop(chan, u->nick, 0);

			if (debug)
				log("SJOIN: '%s' added to channel %s", users, chan);
		}

		if (cptr)
			users = cptr;
		else
			*users = 0;

	}

#ifdef BAHAMUT
	if (valid_user && strchr(source, '.'))
#else /* UNREAL */
	if (valid_user && !c)
#endif
		c_mode(NULL, 5, mv);


	if (mvfree) {
		free(mv[0]); free(mv[2]);
		if (mvfree>3)
			free(mv[3]);
		if (mvfree>4)
			free(mv[4]);
		if (mvfree>5)
			free(mv[5]);
	}


}

/***********************************************************************************************/

static Channel *add_channel_node(const char *chan, time_t stamp)
{
	Channel *c = scalloc(sizeof(Channel), 1);
	ChannelNode *cn, *newnode = scalloc(sizeof(ChannelNode), 1);
	ChanInfo *ci = findchaninfo(chan);
	int value=0, alpha=(strlen(chan)>1) ? toupper(chan[1]) & 0x000000FF : 0;

	if (ci)
		monitored++;

	if (debug)
		log("Creating new channel %s via sjoin", chan);


	if (!cburst) {                     /* have not processed channel burst yet */
		ChannelNode *burstchan = scalloc(sizeof(ChannelNode), 1);

		if (!chanlist)
			chanlist = burstchan;
		else
			end->right = burstchan;

		burstchan->c = c;
		end = burstchan;
	}

	if (strlen(chan) > 2)
		value = toupper(chan[2]) & 0x000000FF;
	if (strlen(chan) > 3)
		value += toupper(chan[3]) & 0x000000FF;

	newnode->c = c;
	newnode->c->node = newnode;
	newnode->value = value;

	strncpy(c->name, chan, sizeof(c->name));
	c->bans = NULL; c->users = NULL;
	c->stamp = stamp;
	cchans++;

	if (cchans > maxchans)
		maxchans = cchans;

	if (cburst && ci && (ci->flags & CS_KEEPTOPIC) && ci->topic && ci->topic_setter && ci->topic_time)
		send_cmd(s_ChanServ, "TOPIC %s %s %lu :%s",
		ci->name, ci->topic_setter, ci->topic_time, ci->topic);


	if (!(cn = channelnodes[alpha])) {
		channelnodes[alpha] = scalloc(sizeof(ChannelNode), 1);
		channelnodes[alpha]->value = 150;
		channelnodes[alpha]->c = scalloc(sizeof(Channel), 1);
		strncpy(channelnodes[alpha]->c->name, "*", 1);
		if (value < 150)
			channelnodes[alpha]->left = newnode;
		else
			channelnodes[alpha]->right = newnode;
		newnode->back = channelnodes[alpha];
		return c;
	}

	while (cn != NULL) {
		if (value < cn->value) {
			if (cn->left)
				cn = cn->left;
			else {
				cn->left = newnode;
				break;
			}
		} else if (value >= cn->value) {
			if (cn->right)
				cn = cn->right;
			else {
				cn->right = newnode;
				break;
			}
		}
	}
	newnode->back = cn;
	return c;
}

/***********************************************************************************************/


void chan_adduser(const char *chan, User *client, time_t stamp)
{
	Channel *c;
	struct ChanUsers *cu = scalloc(sizeof(struct ChanUsers), 1);
	struct UserChans *cl = scalloc(sizeof(struct UserChans), 1);

	if (!(c = findchan(chan)))     /* new channel, create it */
		c = add_channel_node(chan, stamp);


				/* Add user to channel's user list */
	cu->next = c->users;
	cu->prev = NULL;
	if (c->users)
		c->users->prev = cu;
	c->users = cu;
	cu->user = client;
				/* Add channel to user's channel list */

	cl->next = client->chans;
	cl->prev = NULL;
	if (client->chans)
		client->chans->prev = cl;
	client->chans = cl;
	cl->chan = c;


}

/***********************************************************************************************/

/* type 0 = op, 1 = voice, 2 = half */

void chan_addop(const char *chan, const char *nick, int type)
{
	Channel *c = findchan(chan);
	User *u = finduser(nick);
	struct ChanUsers *co;
	char mode='o';

	if (type==1)
		mode='v';
	else if (type==2)
		mode='h';


	if (!c) {
		if (debug)
			log("MODE: Can't set modes for nonexistent channel '%s'", chan);


	} else if (!u) {
		if (debug)
			log("MODE: Can't set nonexistent user '%s' +%c on chan %s", nick, mode, c->name);

#ifdef BAHAMUT
	} else if ((!type && !is_chanop(nick, c->name)) || ((type==1) && !is_voiced(nick, c->name))) {
#else
	} else if ((!type && !is_chanop(nick, c->name)) || ((type==1) && !is_voiced(nick, c->name))
		|| ((type==2) && !is_halfop(nick, c->name))) {
#endif

		if (debug)
			log("MODE: %s +%c %s", c->name, mode, nick);

		co = scalloc(sizeof(struct ChanUsers), 1);


		if (!type) {
			co->next = c->ops;
			co->prev = NULL;
			if (c->ops)
				c->ops->prev = co;
			c->ops = co;
			co->user = u;
		} else if (type==1) {
			co->next = c->voices;
			co->prev = NULL;
			if (c->voices)
				c->voices->prev = co;
			c->voices = co;
			co->user = u;
#ifdef HAS_HALFOP
		} else if (type==2) {
			co->next = c->halfs;
			co->prev = NULL;
			if (c->halfs)
				c->halfs->prev = co;
			c->halfs = co;
			co->user = u;
#endif
		}
	}

}

/***********************************************************************************************/

/* If voice is true, we're deleting a voice, not an op. */

void chan_delop(const char *chan, const char *nick, int type)
{
	Channel *c = findchan(chan);
	User *u = finduser(nick);
	struct ChanUsers *co;
	char mode='o';




	if (type==1)
		mode='v';
	else if (type==2)
		mode='h';


	if (!c) {
		if (debug)
			log("MODE: Can't set modes for nonexistent channel '%s'", chan);

	} else if (!u) {
		if (debug)
			log("MODE: Can't set nonexistent user '%s' -%c on chan %s", nick, mode, c->name);


#ifdef BAHAMUT
	} else if ((!type && is_chanop(nick, c->name)) || ((type==1) && is_voiced(nick, c->name))) {
#else
	} else if ((!type && is_chanop(nick, c->name)) || ((type==1) && is_voiced(nick, c->name))
		|| ((type==2) && is_halfop(nick, c->name))) {
#endif

		if (!type) {
			for (co = c->ops; co; co = co->next) {
				if (sstrcmp(co->user->nick, nick) == 0) {
					if (debug)
						log("MODE: %s -o %s", c->name, nick);
					if (co->next)
						co->next->prev = co->prev;
					if (co->prev)
						co->prev->next = co->next;
					else
						c->ops = co->next;
					free(co);
					break;
				}
			}
		} else if (type==1) {
			for (co = c->voices; co; co = co->next) {
				if (sstrcmp(co->user->nick, nick) == 0) {
					if (debug)
						log("MODE: %s -v %s", c->name, nick);
					if (co->next)
						co->next->prev = co->prev;
					if (co->prev)
						co->prev->next = co->next;
					else
						c->voices = co->next;
					free(co);
					break;

				}
			}
#ifdef HAS_HALFOP
		} else if (type==2) {
			for (co = c->halfs; co; co = co->next) {
				if (sstrcmp(co->user->nick, nick) == 0) {
					if (debug)
						log("MODE: %s -h %s", c->name, nick);
					if (co->next)
						co->next->prev = co->prev;
					if (co->prev)
						co->prev->next = co->next;
					else
						c->halfs = co->next;
					free(co);
					break;

				}
			}
#endif

		}
	}

}


/***********************************************************************************************/

static void check_trimbans(Channel *c)
{
	ChanInfo *ci;
	struct ChanBans *cb, *oldest;
	char *buf[3];
	int times=0, bans=0;


	if (!c || !c->bans || (c->activebans < TRIMBANS))
		return;


	if (!(ci = findchaninfo(c->name)) || !(ci->flags & CS_TRIMBAN))
		return;

	while(c->bans && times < TRIMBANS_REMOVE) {
		oldest = c->bans;
		for (cb = c->bans; cb; cb = cb->next) {
			if (oldest->btime >= cb->btime)
				oldest = cb;
		}

		if (debug)
			log("(trimbans) MODE: %s -b %s", c->name, oldest->mask);


		buf[bans++] = sstrdup(oldest->mask);
		if (bans==4) {
			send_cmd(s_ChanServ, "MODE %s -bbbb %s %s %s %s",
				c->name, buf[0], buf[1], buf[2], buf[3]);
			free(buf[0]); free(buf[1]); free(buf[2]); free(buf[3]);
			bans=0;
		}

		if (oldest->next)
			oldest->next->prev = oldest->prev;
		if (oldest->prev)
			oldest->prev->next = oldest->next;
		else
			c->bans = oldest->next;
		free(oldest->mask);free(oldest);
		c->activebans--;
		times++;
	}
	if (bans) {
		send_cmd(s_ChanServ, "MODE %s -b%s%s %s %s %s", c->name,
			bans>1 ? "b" : "", bans>2 ? "b" : "", buf[0],
			bans>1 ? buf[1] : "", bans>2 ? buf[2] : "");
			free(buf[0]);
			if (bans>1)
				free(buf[1]);
			if (bans>2)
				free(buf[2]);
	}
}

/***********************************************************************************************/


/* If setban is false, only add the ban to memory (set by another client/server).
   If setban is true, add the ban to memory and have chanserv set the mode (ie: akick detection) */


void chan_addban(const char *chan, const char *mask, int setban)
{
	Channel *c = findchan(chan);
	struct ChanBans *cb;


	if (!c) {
		if (debug)
			log("MODE: Can't set modes for nonexistent channel '%s'", chan);

	}

	for (cb = c->bans; cb; cb = cb->next) {
		if (sstrcmp(cb->mask, mask) == 0)
			return;
	}


	if (debug)
		log("MODE: %s +b %s", c->name, mask);

	if (setban)
		send_cmd(s_ChanServ, "MODE %s +b :%s", chan, mask);

	cb = scalloc(sizeof(struct ChanBans), 1);
	cb->next = c->bans;
	cb->prev = NULL;
	if (c->bans)
		c->bans->prev = cb;
	c->bans = cb;
	cb->mask = sstrdup(mask);
	c->activebans++;
	cb->btime = cburst ? time(NULL) : time(NULL) - c->activebans;

/* Since ircd during synch sends the newest bans first, subtracting second from each ban's time will
 * reverse the list. This is much easier than reverse tokenization and will have no detrimental
 * effect on services since the ban time is only used for organization of trimbans. */

	if (cburst)
		check_trimbans(c);


}

/***********************************************************************************************/

void chan_delban(const char *chan, const char *mask)
{
	Channel *c = findchan(chan);
	struct ChanBans *cb;

	if (!c) {
		if (debug)
			log("MODE: Can't set modes for nonexistent channel '%s'", chan);

	}

	for (cb = c->bans; cb; cb = cb->next) {
		if (sstrcmp(cb->mask, mask) == 0) {
			if (debug)
				log("MODE: %s -b %s", c->name, mask);
			if (cb->next)
				cb->next->prev = cb->prev;
			if (cb->prev)
				cb->prev->next = cb->next;
			else
				c->bans = cb->next;
			free(cb->mask);	free(cb);
			c->activebans--;
			break;
		}
	}

}


/***********************************************************************************************/

Channel *findchan(const char *chan)
{
	ChannelNode *cn;
	int value=0, alpha=(strlen(chan)>1) ? toupper(chan[1]) & 0x000000FF : 0;

	if (!(cn = channelnodes[alpha]))
		return NULL;
	if (strlen(chan) > 2)
		value = toupper(chan[2]) & 0x000000FF;
	if (strlen(chan) > 3)
		value += toupper(chan[3]) & 0x000000FF;



	while (cn != NULL) {
		if ((value == cn->value) && sstrcmp(cn->c->name, chan) == 0)
			return cn->c;

		cn = (value<cn->value) ? cn->left : cn->right;
	}
	return NULL;
}



/***********************************************************************************************/

ChanInfo *findchaninfo(const char *chan)
{
	ChanNode *cn;
	int value=0, alpha = toupper(*chan) & 0x000000FF;

	if ((alpha > 256) || (alpha < 0) || !(cn = channodes[alpha]))
		return NULL;


	if (strlen(chan) > 2)
		value = toupper(chan[2]) & 0x000000FF;
	if (strlen(chan) > 3)
		value += toupper(chan[3]) & 0x000000FF;


	while (cn != NULL) {
		if ((value == cn->value) && sstrcmp(chan, cn->ci->name) == 0)
                        return cn->ci;
                cn = (value<cn->value) ? cn->left : cn->right;
	}
	return NULL;
}

/***********************************************************************************************/

int in_chan(const char *nick, const char *chan)
{
	Channel *c = findchan(chan);
	struct ChanUsers *cu;

	if (!c)
		return 0;

	for (cu = c->users; cu; cu = cu->next) {
		if (!sstrcmp(cu->user->nick, nick))
			return 1;
	}

	return 0;

}


/***********************************************************************************************/



static int is_mode(const char *nick, const char *chan, int type)
{
	Channel *c = findchan(chan);
	struct ChanUsers *cu, *cuptr;

	if (!c)
		return 0;


	if (!type)
		cuptr = c->ops;
	else if (type==1)
		cuptr = c->voices;
#ifdef HAS_HALFOP
	else if (type==2)
		cuptr = c->halfs;
#endif
	else
		return 0;


	for (cu = cuptr; cu; cu = cu->next) {
		if (!sstrcmp(cu->user->nick, nick))
			return 1;
	}
	return 0;
}



/***********************************************************************************************/

void delete_chan(Channel *c)
{
	ChannelNode *cn = c->node;
	struct ChanBans *cb = c->bans, *cb2;

	if (debug)
		log("Deleting channel %s - last user exited", c->name);


	if (findchaninfo(c->name))
		monitored--;

	if (c->chankey)
		free(c->chankey);

	while (cb) {
		cb2 = cb->next;
		free(cb->mask); free(cb);
		cb = cb2;
	}

	if (cn->back) {
		if (cn->back->right == cn)
			cn->back->right = cn->right;
		else
			cn->back->left = cn->right;

		if (cn->right)
			cn->right->back = cn->back;
	}


	if (cn->left) {   /* retree */
		ChannelNode *nptr = cn->back, *movenode = cn->left;

		while (nptr != NULL) {
			if (movenode->value < nptr->value) {
				if (nptr->left)
					nptr = nptr->left;
				else {
					nptr->left = movenode;
					break;
				}
			} else if (movenode->value >= nptr->value) {
				if (nptr->right)
					nptr = nptr->right;
				else {
					nptr->right = movenode;
					break;
				}
			}
		}
		movenode->back = nptr;
	}
	free(c); free(cn);
	cchans--;

}





/***********************************************************************************************/


int is_chanop(const char *nick, const char *chan) { return is_mode(nick, chan, 0); }
int is_voiced(const char *nick, const char *chan) { return is_mode(nick, chan, 1); }
int is_halfop(const char *nick, const char *chan) { return is_mode(nick, chan, 2); }


/***********************************************************************************************/

void chan_deluser(User *client, Channel *c)
{
	struct ChanUsers *cu;
	struct UserChans *uc;


	if (!client || !c)
		return;


	for (uc = client->chans; uc; uc = uc->next) {
		if (uc->chan == c) {
			if (uc->next)
				uc->next->prev = uc->prev;
			if (uc->prev)
				uc->prev->next = uc->next;
			else
				client->chans = uc->next;
			free(uc);
			break;
		}
	}


	for (cu = c->users; cu; cu = cu->next) {
		if (cu->user == client) {
			if (cu->next)
				cu->next->prev = cu->prev;
			if (cu->prev)
				cu->prev->next = cu->next;
			else
				c->users = cu->next;
			free(cu);
			break;
		}
	}


	for (cu = c->ops; cu; cu = cu->next) {
		if (cu->user == client) {
			if (cu->next)
				cu->next->prev = cu->prev;
			if (cu->prev)
				cu->prev->next = cu->next;
			else
				c->ops = cu->next;
			free(cu);
			break;
		}
	}

	for (cu = c->voices; cu; cu = cu->next) {
		if (cu->user == client) {
			if (cu->next)
				cu->next->prev = cu->prev;
			if (cu->prev)
				cu->prev->next = cu->next;
			else
				c->voices = cu->next;
			free(cu);
			break;
		}
	}


#ifdef HAS_HALFOP
	for (cu = c->halfs; cu; cu = cu->next) {
		if (cu->user == client) {
			if (cu->next)
				cu->next->prev = cu->prev;
			if (cu->prev)
				cu->prev->next = cu->next;
			else
				c->halfs = cu->next;
			free(cu);
			break;
		}
	}
#endif


	if (!c->users)
		delete_chan(c);

}

/***********************************************************************************************/

void do_part(char *source, int ac, char **av)
{
	Channel *c;
	User *u;
	const char *chan = av[0];

	if (!(u = finduser(source))) {
		if (debug)
			log("PART: Nonexistent user '%s' from channel '%s'", source, chan);

	} else if (!(c = findchan(chan))) {
		if (debug)
			log("PART: User '%s' from nonexistent channel '%s'", source, chan);

	} else {
		if (debug)
			log("PART: User '%s' from channel '%s'", source, chan);


		chan_deluser(u,c);


	}

}

/***********************************************************************************************/

/* When a client leaves the network in any way this will make sure all it's UserChans are deleted. */

void delete_uchans(User *u)
{
	struct UserChans *cl = u->chans, *next;

	while (cl) {
		next = cl->next;
		chan_deluser(u, cl->chan);
		cl = next;
	}
}

/***********************************************************************************************/

void do_kick(char *source, int ac, char **av)
{
	Channel *c;
	User *u;

	if (ac != 3)
		return;


	if (!(u = finduser(av[1]))) {
		if (debug)
			log("KICK: %s kicked non-existent user %s from channel %s",
				source, av[1], av[0]);

	} else if (!(c = findchan(av[0]))) {
		if (debug)
			log("KICK: %s kicked user %s from non-existent channel %s",
				source, av[1], av[0]);

	} else {

		chan_deluser(u,c);

	}
}

/***********************************************************************************************/

void process_channel_burst()
{
	ChannelNode *cn = chanlist, *c2;
	Channel *c;
	struct ChanUsers *cu, *cu2;
	ChanInfo *ci;
	int updatelast=0;

	while (cn) {
		c = cn->c;
		c2 = cn->right;
		if ((ci = findchaninfo(c->name)))
			check_modes(c, ci);

		updatelast=0;





		if (ci && (ci->flags & CS_KEEPTOPIC) && ci->topic && ci->topic_setter && ci->topic_time)
			send_cmd(s_ChanServ, "TOPIC %s %s %lu :%s",
				ci->name, ci->topic_setter, ci->topic_time, ci->topic);


		cu = c->ops;
		while (cu) {
			cu2 = cu->next;
			if (!check_valid_op(cu->user, c->name)) {
				if (cu->next)
					cu->next->prev = cu->prev;
				if (cu->prev)
					cu->prev->next = cu->next;
				else
					c->ops = cu->next;
				free(cu);
			} else {
#ifdef UNREAL
				check_special_op(cu->user, c->name);
#endif
				updatelast=1;
			}
			cu = cu2;
		}


		cu = c->users;
		while (cu) {
			int acc=0;
			cu2 = cu->next;

			if (check_should_kick(cu->user, c->name)) {
				chan_deluser(cu->user, c);

			} else if (!is_chanop(cu->user->nick, c->name)) {
				if ((acc=check_should_op(cu->user, c->name))) {
					chan_addop(c->name, cu->user->nick, 
					acc==3 ? 1 : acc==4 ? 2 : 0);
					updatelast=1;
				}
			}
			cu = cu2;
		}

		if (ci && ci->flags & CS_TRIMBAN) {
			while (c->activebans > TRIMBANS)
				check_trimbans(c);
		}



		if (ci && updatelast)
			send_sql("UPDATE chan SET lasttime=%d WHERE id = %d", time(NULL), ci->id);

		free(cn);
		cn = c2;
	}
	chanlist=NULL;
	end=NULL;
	cburst=1;
}

/***********************************************************************************************/

ChanInfo *add_ci_chan(const char *chan, const char *founder, long flags)
{
	ChanNode *cn, *newnode = scalloc(sizeof(ChanNode), 1);
	ChanInfo *ci = scalloc(sizeof(ChanInfo), 1);
	int alpha = toupper(*chan) & 0x000000FF;
	int value=0;

	if (strlen(chan) > 2)
		value = toupper(chan[2]) & 0x000000FF;
	if (strlen(chan) > 3)
		value += toupper(chan[3]) & 0x000000FF;



	strncpy(ci->name, chan, sizeof(ci->name));
	strncpy(ci->founder, founder, sizeof(ci->founder));
	ci->flags = flags; ci->welcome = NULL; ci->topiclock = 0; ci->mlevel=5;

	newnode->ci = ci;
	newnode->value = value;


	if (!(cn = channodes[alpha])) {
		channodes[alpha] = scalloc(sizeof(ChanNode), 1);
		channodes[alpha]->value = 150;
		channodes[alpha]->ci = scalloc(sizeof(ChanNode), 1);
		strncpy(channodes[alpha]->ci->name, "*", 1);
		if (value < 150)
			channodes[alpha]->left = newnode;
		else
			channodes[alpha]->right = newnode;
		return ci;
	}
	while (cn != NULL) {
		if (value < cn->value) {
			if (cn->left)
				cn = cn->left;
			else {
				cn->left = newnode;
				break;
			}
		} else if (value >= cn->value) {
			if (cn->right)
				cn = cn->right;
			else {
				cn->right = newnode;
				break;
			}
		}
	}

	return ci;
}

/***********************************************************************************************/

static void retree_channode(ChanNode *cn, ChanNode *nptr)
{
	while (nptr != NULL) {
		if (cn->value < nptr->value) {
			if (nptr->left)
				nptr = nptr->left;
			else {
				nptr->left = cn;
				break;
			}
		} else if (cn->value >= nptr->value) {
			if (nptr->right)
				nptr = nptr->right;
			else {
				nptr->right = cn;
				break;
			}
		}
	}
}

/***********************************************************************************************/

void del_ci_chan(const char *chan)
{
	ChanNode *cn, *last;
	int value=0, alpha=toupper(*chan) & 0x000000FF;


	if (strlen(chan) > 2)
		value = toupper(chan[2]) & 0x000000FF;
	if (strlen(chan) > 3)
		value += toupper(chan[3]) & 0x000000FF;


	last = cn = channodes[alpha];

	while (cn != NULL) {
		if ((value == cn->value) && sstrcmp(chan, cn->ci->name) == 0) {
			ChanInfo *ci = cn->ci;
			struct AccessList *a, *a2;
			struct u_founderlist *fl, *fl2;
			User *u;
			int ulist=0;

			freeze(ci->name, 0);
			official_list(ci, DEL);

			a = ci->entries;

			while (a) {
				a2 = a->next;
				if (a->reason)
					free(a->reason);
				free(a->entry);free(a->instater);free(a);
				a = a2;
			}


			for (ulist=0; ulist < 256; ulist++) {
				for (u = userlist[ulist]; u; u = u->next) {
					fl = u->fchans;
					while (fl) {
						fl2 = fl->next;
						if (fl->chan == ci) {
							if (fl->prev)
								fl->prev->next = fl->next;
							else
								u->fchans = fl->next;
							if (fl->next)
								fl->next->prev = fl->prev;
							free(fl);
						}
						fl = fl2;
					}
				}
			}


			if (ci->topic)
				free(ci->topic);
			if (ci->welcome)
				free(ci->welcome);
			free(ci);
			if (findchan(chan))
				monitored--;


			/* remove node attached to chaninfo structure */

			if (last->right == cn)
				last->right = cn->right;
			else if (last->left == cn)
				last->left = cn->right;
			if (cn->left)
				retree_channode(cn->left, last);
			free(cn);

			return;

		}
		last = cn;
		cn = (value<cn->value) ? cn->left : cn->right;
	}


	if (debug)
		log("Tried to delete channel '%s' that is in database but not in memory!", chan);


}

/***********************************************************************************************/

void load_chaninfo()
{
	ChanNode *cn, *cn2;
	ChanInfo *ci;
	struct AccessList *a;
	unsigned long num_rows = 
		sql_select("name,topiclock,flags,welcome,founder,topic,tsetter,ttime,mlevel,id,fid,pass FROM chan");
	unsigned long i;

	for (i=0; i < num_rows; i++) {
		row = fetch(res);
		ci = add_ci_chan(row[0], row[4], atol(row[2]));

			cn = scalloc(sizeof(ChanNode), 1); /* Temporary node list for below */
			cn->ci = ci;
			cn->right = ci_list;
			ci_list = cn;

		ci->topiclock=atoi(row[1]);
		ci->id=atol(row[9]);
		ci->fid=atol(row[10]);
		ci->entries = NULL;
		if (row[3])
			ci->welcome=sstrdup(row[3]);

		if (ci->flags & CS_KEEPTOPIC) {
			if (row[5])
				ci->topic = sstrdup(row[5]);
			if (row[6])
				strncpy(ci->topic_setter, row[6], sizeof(ci->topic_setter));
			if (row[7])
				ci->topic_time = atol(row[7]);
		}
		ci->mlevel = atoi(row[8]);
		if (ci->flags & CS_OFFICIAL)
			official_list(ci, ADD);
#ifdef USE_ENCRYPTION
		if (!(ci->flags & CS_ENCRYPTED)) {
			char passwd[PASSMAX];

			if (!encryptm(row[11], strlen(row[11]), passwd, PASSMAX, 1)) {
				ci->flags |= CS_ENCRYPTED;
				send_sql("UPDATE chan SET pass = '%s', flags = %d WHERE id = %d",
					passwd, ci->flags, ci->id);
			} else
				fatal_error("Could not encrypt password for chan '%s'", ci->name);

		}
#else  /* NOT ENCRYPTED */
		if ((ci->flags > 0) && ci->flags & CS_ENCRYPTED)
			fatal_error("Chan '%s' has encrypted password and services are not in "
				"encryption mode!", ci->name);
#endif

	}

	cn = ci_list;
	while (cn != NULL) {
		cn2 = cn->right;
		num_rows = sql_select("* FROM caccess WHERE chan = %d", cn->ci->id);

		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			a = scalloc(sizeof(struct AccessList), 1);
			a->next = cn->ci->entries;
			a->prev = NULL;
			if (cn->ci->entries)
				cn->ci->entries->prev = a;
			cn->ci->entries = a;
			a->entry = sstrdup(row[1]);
			a->instater = sstrdup(row[3]);
			a->level = atoi(row[2]);
			a->when = atol(row[4]);
			a->reason = row[5] ? sstrdup(row[5]) : NULL;
		}
		for (a = cn->ci->entries; a; a = a->next) {
			if (a->level == -2) {
				if (sql_select("id FROM nick WHERE name = '%s'", a->entry)) {
					row = fetch(res);
					a->nickid = strtoul(row[0], NULL, 10);
				}
			}
		}
		cn = cn2;
	}
}

/***********************************************************************************************/

void do_topic(char *source, int ac, char **av)
{
	Channel *c = findchan(av[0]);
	ChanInfo *ci = findchaninfo(av[0]);

	if (!ci)
		return;


	if (!c) {
		if (debug)
			log("DEBUG: TOPIC from non-existent channel %s.", av[0]);
		return;
	}


	if (check_topiclock(ci, source)) {
		send_cmd(s_ChanServ, "TOPIC %s %s %lu :%s",
			av[0], ci->topic_setter, ci->topic_time, ci->topic);
		return;
	}

	if (ci->topic)
		free(ci->topic);

	ci->topic = sstrdup(av[3]);
	ci->topic_time = time(NULL);
	strncpy(ci->topic_setter, av[1], sizeof(ci->topic_setter));

	if (ci->flags & CS_KEEPTOPIC)
		send_sql("UPDATE chan SET topic = '%s', tsetter = '%s', ttime = %d WHERE id = %d",
			av[3], ci->topic_setter, time(NULL), ci->id);
}

/***********************************************************************************************/

int chan_closed(const char *chan)
{
	CloseChans *cc;

	for (cc = closelist; cc; cc = cc->next) {
		if (sstrcmp(cc->chan, chan) == 0)
			return 1;
	}
	return 0;
}

/***********************************************************************************************/


/*
 * If opt equals 0 then delete the node.
 * If opt equals 1 then add the node.
 * If opt equals 3 then list the next node after the given ci pointer.
 * If opt equals 3 and the ci pointer is NULL, then start with the first.
 */


ChanInfo *official_list(ChanInfo *ci, int opt)
{
	ChanNode *official;

	if (opt == 3) {
		if (!ci)
			return officials ? officials->ci : NULL;

		for (official = officials; official; official = official->right) {
			if (official->ci == ci) {
				if (official->right)
					return official->right->ci;
			}
		}
		return NULL;
	}

	for (official = officials; official; official = official->right) {
		if (official->ci == ci) {
			if (opt)
				return ci;
			if (official->right)
				official->right->left = official->left;
			if (official->left)
				official->left->right = official->right;
			else
				officials = official->right;
			free(official);
			return NULL;
		}
	}
	if (!opt)
		return NULL;

	official = scalloc(sizeof(ChanNode), 1);
	official->ci = ci;
	official->right = officials;
	if (officials)
		officials->left = official;
	officials = official;

	return official->ci;
}

/***********************************************************************************************/
/***********************************************************************************************/

