/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */

#include "services.h"
#include "log.h"

static char *pidfile = "services.pid";



/********************************************************************/

static inline void remove_pidfile() { remove(pidfile); }

static void write_pidfile()
{
	FILE *pid;

	pid = fopen(pidfile, "w");
	if (pid) {
		fprintf(pid, "%d\n", getpid());
		fclose(pid);
		atexit(remove_pidfile);
	} else
		log("WARNING: couldn't write to PID file %s", pidfile);
}


/********************************************************************/

int init()
{
	uid_t   uid, euid;
	int     i, passed=1;

	fprintf(stderr, "Starting up SirvNET Services..\n");

	uid = getuid(); euid = geteuid();
	if ((uid != euid) && !euid) {
		fprintf(stderr, "Do not run services as root..\n");
		return 0;
	} else
		fprintf(stderr, "Passed(%d): id check\n", passed++);


	if (chdir(BINDEST) < 0) {
		fprintf(stderr, "ERROR: can't chdir into %s: (%s)\n",
  		   BINDEST, strerror(errno));
		return 0;
	}
	if (chdir(DATDEST) < 0) {
		fprintf(stderr, "ERROR: can't chdir into %s (%s)\n",
		  DATDEST, strerror(errno));
		return 0;
	}
	if (chdir("./logs") < 0)
		system("mkdir ./logs");

	else if (chdir("..") < 0) {
		fprintf(stderr, "Directory structure error..aborting\n");
		return 0;
	}
	fprintf(stderr, "Passed(%d): directory structure\n", passed++);


	if (!READ_ONLY) {
		if (!open_logs())
			return 0; /* open_logs() will give a fail reason */
		else
			fprintf(stderr, "Passed(%d): log descriptors\n", passed++);
	} else
		fprintf(stderr, "Passed(%d): bypassing logfiles - readonly set\n",passed++);

	if (!init_conf())
		return 0;  /* init_conf() will give a fail reason */
	else {
		if (CONNECT != 1) {
			i=0;
			fprintf(stderr, "      (4): Allow IP(s): ");
			while (allow_ips[i++])
				fprintf(stderr, "[%i] %s ", i, allow_ips[i-1]);
				fprintf(stderr, "\n");
		        }
		fprintf(stderr, "Passed(%d): configuration\n", passed++);
	}

	if (!connect_sql())    /* function will relay it's own fail messages */
		fatal_error("Sirv Start-up requires to be connected to an SQL database server!%s",
		nofork==1 ? "Run again with -nofork option for further information and fail message." : "");


	fprintf(stderr, "Loading database information...\n");
	load_closures();
	load_freezes();
	load_akills();
	load_triggers();
	load_ignores();
	load_nickinfo();
	load_chaninfo();
	load_news();
	fprintf(stderr, "Passed(%d): SQL Database\n", passed++);


	if (!nofork) {
		if ((i = fork()) < 0) {
			fprintf(stderr, "\nERROR: couldn't fork: (%d)\n", errno);
			return 0;
		} else if (i != 0) {
		 	fprintf(stderr, "Passed(%d): fork()\n",passed); 

	fprintf(stderr, "Connect behaviour: %s%s%s%s%s\n",
		CONNECT==0 ? "Listen for connections" : "",
		CONNECT==1 ? "Will attempt to connect once" : "",
		CONNECT==2 ? "Listen for connections after connection attempt" : "",
		RECONNECT==0 ? " and will not attempt to reconnect." : "",
		RECONNECT==1 ? " and reconnect if link broken." : "");


			fprintf(stderr, "[Sirv running in background] [Modes: %s%s%s%s%s] [Dbase: %s]\n",
				use_services ? "services" : "stats",
				(use_services && use_stats) ? ", " : "",
				(use_services && use_stats) ? "stats" : "",
				READ_ONLY ? ", read-only" : "",
				debug ? ", debug" : "",
				use_sql ? "SQL" : "conventional");

			fprintf(stderr, "This will be the last process message - "
			  "if services do not run or something unexpected happens use: ./services -nofork\n");

			exit(0);
		}
	} else
		fprintf(stderr, "bypassing fork()\n");



	if ((i = setpgid(0,0)) == -1) {
		fprintf(stderr, "ERROR: can't set child id: (%d)\n", errno);
		return 0;
	}

	log("SirvNET Services starting up [Modes: %s%s%s%s%s] [Dbase: %s]",
		use_services ? "services" : "stats",
		(use_services && use_stats) ? ", " : "",
		(use_services && use_stats) ? "stats" : "",
		READ_ONLY ? ", read-only" : "",
		debug ? ", debug" : "",
		use_sql ? "SQL" : "conventional");

	write_pidfile();


   return 1;
}

/********************************************************************/

