/*  SirvNET Services' main routines and functions
 *
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */


#include "services.h"
#include "log.h"
#include "timeout.h"

int debug_traffic = 0;
int debug = 0;
int nofork = 0;
int sshutdown = -1;  /* -1 starting up, 0 running, 1 shutting down, 2 restart, -2 convert 2.9, -3 convert 1.7  */
int sql_alive = 0;   /* is the sql database server up (1) or down (0) */
int uburst=0;         /* have we processed user burst? */
int cburst=0;        /*    "   "     "     chan burst? */
time_t start_time;
int skeleton_mode = 0;  /* 0 = dormant, 1 = active */


char inbuf[512];


#ifdef USE_SERVICES
  int use_services = 1;
#else
  int use_services = 0;
#endif
#ifdef USE_STATS
  int use_stats = 1;
#else
  int use_stats = 0;
#endif
#ifdef USE_SQL
  int use_sql = 1;
#else
  int use_sql = 0;
#endif



fd_set master, smtp_fdlist;
fd_set read_fds, smtp_read_fds;
int fdmax, smtp_fdmax;

/************************************************************************/
/************************************************************************/

void fatal_error(const char *fmt,...)
{
	va_list args;
	char fbuf[1024];

	va_start(args, fmt);
	vsnprintf(fbuf, sizeof(fbuf), fmt, args);


	if (sshutdown == 0)
		globops(NULL, "Services exiting on fatal error: %s", fbuf);

	if (logfile)
		log("FATAL ERROR: %s", fbuf);
	if (sshutdown == -1 && !nofork) {
		fprintf(stderr, "ERROR: %s\n", fbuf);
		fprintf(stderr, "Services not started..\n");
	}
	exit(1);
}


/************************************************************************/

static void show_help()
{
	fprintf(stderr, "usage:  services [option]\n\n");
	fprintf(stderr, "Options:\n\n");
#ifdef CONVERT_29_ANOPE
	fprintf(stderr, "   -convert17    convert anope 1.7 db's into SQL server\n");
	fprintf(stderr, "   -convert29    convert sirv 2.9 db's into SQL server\n");
#endif
	fprintf(stderr, "   -debug        load into debug mode\n");
	fprintf(stderr, "   -debug2       extensive debug w/ traffic\n");
	fprintf(stderr, "   -nofork       do not fork() into background\n");
	fprintf(stderr, "   -readonly     no disk/database writing\n\n");
	fprintf(stderr, "END OF OPTION HELP\n");
}

/************************************************************************/


static int parse_options(int ac, char **av)
{
        int i;
        char *opt;

        for (i=1; i < ac; i++) {
                opt = av[i];
                if (*opt == '-') {
                        ++opt;
                        if (sstrcmp(opt, "nofork") == 0)
                                nofork = 1;
			else if (!sstrcmp(opt, "help") || !sstrcmp(opt, "-help")
				|| !sstrcmp(opt, "?") || !sstrcmp(opt, "-?")) {
				show_help();
				return 0;
			} else if (sstrcmp(opt, "readonly") == 0) {
				READ_ONLY = 2; /* 2 overrides conf */
			} else if (sstrcmp(opt, "debug") == 0) {
				debug = 2; /* 2 overrides conf */
			} else if (sstrcmp(opt, "debug2") == 0) {
				debug_traffic = 2;
#ifdef CONVERT_29_ANOPE
			} else if (sstrcmp(opt, "convert17") == 0) {
				sshutdown = -3; debug=1; nofork=1;

			} else if (sstrcmp(opt, "convert29") == 0) {
				sshutdown = -2; debug=1; nofork=1;
#endif /* CONVERT_29_ANOPE */
			} else {
				fprintf(stderr, "Invalid Option: %s\n", opt);
				show_help();
                                return 0;
                        }
                } else {
                        fprintf(stderr, "Invalid argument\n");
                        return 0;
                }
        }
	return 1;
}

/************************************************************************/


int main(int ac, char **av, char **envp)
{
	time_t routine = time(NULL);
	time_t checklog = time(NULL);
	time_t sqlping = time(NULL);
#ifdef BAHAMUT
	time_t long_update = time(NULL);
#endif
	time_t web_update = time(NULL);
#ifndef SKELETON
	time_t check_update = time(NULL);
	time_t backup_update = time(NULL);
#endif
	int i;
	struct timeval to, s_to;


	if (!parse_options(ac, av)) {
		fprintf(stderr, "Services not started...\n");
		return 0;
	}

	if (!init()) {
		fprintf(stdout, "Initilization failed...terminating\n");
		return 0;
	}

#ifdef CONVERT_29_ANOPE
	if (sshutdown == -2) {  /* convert sirv 2.9 database */
		i = convert29();
		fprintf(stderr, "Sirv 2.9 database conversion %s.\n",
			i ? "successful" : "failed");
		return 0;
	}
	if (sshutdown == -3) {  /* convert anope 1.7 database */
		i = convert17();
		fprintf(stderr, "Anope 1.7 database conversion %s.\n",
			i ? "successful" : "failed");
		return 0;
	}
#endif /* CONVERT_29_ANOPE */


	sshutdown = 0;
	start_time = time(NULL);


	FD_ZERO(&master); FD_ZERO(&read_fds);
	FD_ZERO(&smtp_fdlist); FD_ZERO(&smtp_read_fds);

	if (!serv_start())
		return 0;

	while (!sshutdown) {
		to.tv_sec = 1; to.tv_usec = 0;
		s_to.tv_sec = 0; s_to.tv_usec = 1;

	/* check for servfd for web update since it would be poinltess to check if we are not
	   connected to the irc network to notify anyone. */

		if (WEB_UPDATE && servfd && ((time(NULL) - web_update) >= (WEB_UPDATE_HOURS*60*60))) {
			init_updatesock();
			update_connect();
			web_update = time(NULL);
		}

		if ((time(NULL) - routine) >= TIMEOUT_CHECK) {
			check_timeouts();
			kill_emails();
			expire_akills();
			expire_ignores();
			expire_closures();
			routine = time(NULL);
		}
		if ((time(NULL) - checklog) >= 180) {
			check_logs();
			checklog = time(NULL);
		}
#ifndef SKELETON
		if ((time(NULL) - check_update) >= (time_t)UPDATE_TIMEOUT) {
			int x=0;
			if (DISPLAY_UPDATES)
				globops(NULL, "Running Maintainence...");
			x=expire_pends();
			expire_nicks(x);
			expire_chans();
			expire_memos();
			check_update = time(NULL);
		}
		if ((time(NULL) - backup_update) >= (time_t)(BACKUP_UPDATE*60*60)) {
			backup_routine();
			backup_update = time(NULL);
		}

#endif
#ifdef BAHAMUT
		/* See notes under misc.c for the logic of this routine */
		if ((time(NULL) - long_update) >= 604800) {
			send_akills();
			long_update = time(NULL);
		}
#endif

		if ((time(NULL) - sqlping) >= SQL_CHECK) {
			if (sql_alive) {
				if (!ping_sql()) { /* we've lost connection to sql db */
					sql_alive = 0;
					cripple_pseudo();
				}
			} else { /* let's attempt reconnect */
				if (connect_sql()) {
					sql_alive = 1;
					if (servfd)
						globops(NULL, "Connection to SQL database server "
						"restored. Services returned to full functionality.");
				}
			}
			sqlping = time(NULL);
		}

		if (fdmax > 0) {
			read_fds = master;
			select(fdmax+1, &read_fds, NULL, NULL, &to);
		} else if (!RECONNECT && (CONNECT == 1))
			break;  /* lost connection, let's shutdown */
		else
			sleep(1); /* prevent cpu loop */

		if (servfd && FD_ISSET(servfd, &read_fds)) {
			i = (int)newmsg(servfd, inbuf, sizeof(inbuf));
			if (i > 0)
				process(servfd);
			else
				serv_lost();
		} else if ((servlfd > 0) && FD_ISSET(servlfd, &read_fds)) {
			accept_serv();

		} else if (updatefd && FD_ISSET(updatefd, &read_fds)) {
			i = (int)newmsg(updatefd, inbuf, sizeof(inbuf));
			if (i > 0)
				web_update_check();
			else
				end_web_check();
		}

		if (SMTP_SERVER && (smtp_fdmax > 0)) {  /* anything else must be outgoing SMTP email */
			smtp_read_fds = smtp_fdlist;
			if (select(smtp_fdmax+1, &smtp_read_fds, NULL, NULL, &s_to))
				select_smtp();
		}



	}  /* while !shutdown */

	if (sshutdown == 2) {
		serv_lost();
		chdir(DATDEST);
		execve(BINDESTB, av, envp);
		return 1;
	}

	if (!sshutdown && logfile)
		log("Abnormal shutdown...");

	return 0;
}

/************************************************************************/
/************************************************************************/

