/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */

#include "services.h"
#include "log.h"

/****************************************************************************/

char *operconf_name = "oper.conf";


/****************************************************************************/

int has_access(const char *source, int level)
{
	User *u = finduser(source);

	if (!level)   /* BASIC command available to all */
		return 1;

	else if (!is_oper(source))
		return 0;

	else if (level == 1)  /* OPER command */
		return 1;

	else if (u && u->os_level >= (unsigned int)(level-1))
		return 1;

	return 0;

}


/****************************************************************************/

int oper_conf_break(int pargc, char **pargv)
{
	char buf[64];
	int i, valid=0;

	*buf = 0;

	if (*pargv[0] != '#') {
		strcat(buf, pargv[0]);

		if (pargc != 2)
			fatal_error("Error in %s Option \"%s\"", operconf_name, pargv[0]);

		if (atoi(pargv[1]) < 0 || atoi(pargv[1]) > 5)
			fatal_error("Error in %s Option \"%s\" (Must be between 0-5)",
				pargv[0], operconf_name);

		if ((i=check_oper_command(pargv[0])) > -1) {
			valid=1;
			update_oper_table(i, (atoi(pargv[1])));
		}
		if ((i=check_ns_command(pargv[0])) > -1) {
			valid=1;
			update_ns_table(i, (atoi(pargv[1])));
		}
		if ((i=check_cs_command(pargv[0])) > -1) {
			valid=1;
			update_cs_table(i, (atoi(pargv[1])));
		}
		if ((i=check_ms_command(pargv[0])) > -1) {
			valid=1;
			update_ms_table(i, (atoi(pargv[1])));
		}

		if (!valid)
			fatal_error("Unknown Option \"%s\" in %s", pargv[0], operconf_name);
		return 1;

	} else
		return 1;


}


/****************************************************************************/

void init_operconf() {

	FILE *conf_file;
	char linebuf[1024];
	char *pargv[100];
	char *tmp, conf[512];
	int p;

	snprintf(conf, sizeof(conf), "%s/%s", CONFDIR, operconf_name);
	if ((conf_file = fopen(conf, "rb")) == NULL) {
		log("warning: couldn't open oper permission configuration file = %s",
			operconf_name);
		log("Using defaults for permission levels");
		return;
	}
	while (!feof(conf_file)) {
		if ((fgets(linebuf, 1024, conf_file)) == NULL)
			break;
		p = 0;
		tmp = strtok(linebuf, ": \n\r,");

		if (tmp != NULL) {
			pargv[p++] = tmp;
			while ((tmp = strtok(NULL, ": \n\r,")) != NULL) {
				pargv[p++] = tmp;
				if (p > 100)
					break;
			}
			if (oper_conf_break(p, pargv) == -1)
				fatal_error("Sirv terminated on illegal configuration");
		}
	}
	fclose(conf_file);
	if (sshutdown == -1)
		fprintf(stderr, "      (4): %s\n", operconf_name);
}



/****************************************************************************/
/****************************************************************************/

