/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */


#include "services.h"
#include "timeout.h"
#include "log.h"


struct sockaddr_in accept_sock;

/************************************************************************/
/************************************************************************/


static inline void availability()
{
#ifndef SKELETON
	send_cmd(NULL, "ISON %s %s %s %s %s %s %s %s Guest",
	s_NickServ, s_ChanServ, s_MemoServ, s_HelpServ, s_OperServ,
	s_RootServ, s_Global, s_StatServ);
#else
	send_cmd(NULL, "ISON %s %s", s_OperServ, s_RootServ);
#endif
}


/************************************************************************/

static int check_allow_ip()
{
	int i=0;
	long x;
	char buf[1024];

	x = ntohl(accept_sock.sin_addr.s_addr);

	sprintf(buf, "%d.%d.%d.%d",
		(int)(x>>24) & 0xff, (int)(x>>16) & 0xff,
		(int)(x>>8) & 0xff, (int) (x>>0) & 0xff);

	while (allow_ips[i++]) {
		if (strcmp(allow_ips[i-1], buf) == 0) {
			log("(Services) Incoming connection accepted from: %s", buf);
			return 1;
		}
	}
	log("(Services) Unauthorized connection attempt from: %s", buf);
	return 0;
}


/************************************************************************/

static void serv_recon(Timeout *to)
{
        Timeout *t;

        if (servfd)
                return;

        log("(Services) Attempting to reconnect services to server...");

        if (!CONNECT) /* need to initilize if listen mode with reconnect */
                init_servsock(remote_server, services_port);

        if (serv_connect() == 1) {
                log("(Services) Reconnection to remote server successful");
                FD_SET(servfd, &master);
                if (servfd > fdmax)
                        fdmax = servfd;
		availability();
        } else {
                if (CONNECT != 1) {
                        log("(Services) Reconnect to server failed... re-listening on port %d", SERV_LISTEN_PORT);
                        if (!servlfd)
                                serv_listen();
                        else
                                FD_SET(servlfd, &master);

                        if (servlfd > fdmax)
                                fdmax = servlfd;

                } else {
                        log("(Services) Reconnect to server failed... re-trying in %d seconds", RECON_FREQ);
                        t = add_timeout(RECON_FREQ, serv_recon, 0);
                }
        }
}

/************************************************************************/

int serv_start()
{
	if (CONNECT) {
		Timeout *t;

		init_servsock(remote_server, services_port);
		serv_connect();
		if (servfd > 0) {
			FD_SET(servfd, &master);
			if (servfd > fdmax)
				fdmax = servfd;
			availability();
		} else if (RECONNECT && (CONNECT != 2))
			t = add_timeout(RECON_FREQ, serv_recon, 0);
	}
	if (!CONNECT || ((CONNECT == 2) && !servfd)) {
		if (serv_listen()) {
			if (servlfd > 0) {
				log("(Services) Listening for connection on port %d", SERV_LISTEN_PORT);
				FD_SET(servlfd, &master);
				if (servlfd > fdmax)
					fdmax = servlfd;
			}
		} else
			return 0;
	}

	return 1;
}


/************************************************************************/


void serv_lost()
{

	Timeout *t;

	log("(Services) Lost connection");
	FD_CLR(servfd, &master);
	shutdown(servfd, SHUT_RDWR);
	close(servfd);
	servfd=0;
	if (updatefd)
		fdmax=updatefd;
	else
		fdmax=0;
	uburst=0;cburst=0;pingburst=0;



	if (myhub)  /* otherwise this instance hasn't been connected to any hub yet. */
		squit_server(myhub->name);


	if (RECONNECT) {
		log("(Services) Preparing to re-connect");
		t = add_timeout(RECON_FREQ, serv_recon, 0);
	} else if (CONNECT != 1) {
		if (!servlfd)
			serv_listen();
			log("(Services) Re-listening on port %d", SERV_LISTEN_PORT);
			FD_SET(servlfd, &master);
			if (servlfd > fdmax)
				fdmax = servlfd;
		}
}


/************************************************************************/

void accept_serv()
{
	unsigned int newsock = sizeof(accept_sock);
	servfd = accept(servlfd, (struct sockaddr *)&accept_sock, &newsock);
	if (check_allow_ip()) {
		send_cmd(NULL,  "PASS %s :TS", services_pass);
		send_cmd(NULL, "SERVER %s 1 :%s", servername, services_desc);
		send_cmd(NULL, "SVINFO %d %d 0 :%d", 3, 1, time(NULL));
		send_cmd(NULL, "CAPAB :TS3");
		send_cmd(NULL, "SERVER %s 2 :Statistical Database", statsname);
		FD_SET(servfd, &master);
		FD_CLR(servlfd, &master);
		if (servfd > fdmax)
			fdmax = servfd;
		availability();
	} else {
		send_cmd(NULL, "CONNECTION REFUSED");
		shutdown(servfd, SHUT_RDWR);
		close(servfd);
	}
}


/************************************************************************/
/************************************************************************/

