/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 */


#include "services.h"
#include "log.h"


#define TS_MIN     3
#define TS_MAX     5


int servfd, servlfd, updatefd;

static int serv_initialized = 0, update_initialized=0;

struct sockaddr_in serv_addr, update_addr;
struct hostent *servh, *updateh;

struct sockaddr_in sserv_server;

/****************************************************************************************************/
/****************************************************************************************************/

int serv_listen()
{
	unsigned int len;

	servlfd = socket(AF_INET, SOCK_STREAM, 0);
	if (servlfd < 0) {
		log("(Services) error opening services socket for listening");
		return 0;
	}
	sserv_server.sin_family = AF_INET;
	sserv_server.sin_addr.s_addr = INADDR_ANY;
	sserv_server.sin_port = htons(SERV_LISTEN_PORT);

	if (bind(servlfd, (struct sockaddr *)&sserv_server, sizeof(sserv_server))) {
		log("(Services) error binding services socket for listening");
		return 0;
	}

	len = sizeof(sserv_server);
	if (getsockname(servlfd, (struct sockaddr *)&sserv_server, &len) != 0) {
		log("(Services) error getting socket name");
		return 0;
	}

	listen(servlfd, 100);
	return 1;
}

/****************************************************************************************************/

int init_servsock(const char *rserv, int rport)
{

	if (serv_initialized)
		return 0;
	serv_initialized=1;

	if (!rserv) {
		fatal_error("Invalid remote server");
		return 0;
	} else if (rport < 1024) {
		fatal_error("invalid remote port");
		return 0;
	} else if (!(servh = gethostbyname(rserv))) {
		fatal_error("error in gethostbyname()");
		return 0;
	}

	bzero(&serv_addr, sizeof(serv_addr));
	bcopy(servh->h_addr, (char *)&serv_addr.sin_addr, servh->h_length);
	serv_addr.sin_family = servh->h_addrtype;
	serv_addr.sin_port = htons((unsigned short)rport);
	if (local_ip != NULL)
		serv_addr.sin_addr.s_addr = inet_addr(local_ip);

	return 1;
}

/****************************************************************************************************/

int serv_connect()
{
	servfd = socket(AF_INET, SOCK_STREAM, 0);

	if (servfd < 0) {
		log("(Services) error opening services socket");
		return 0;
	}

	if (local_ip != NULL) {
		if (bind(servfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) != 0)
			log("(Services) WARNING: binding local socket to ip %s failed!", local_ip);
	}

	if (connect(servfd, (struct sockaddr *)&serv_addr,
		sizeof(serv_addr)) < 0) {
		log("(Services) error connecting services to remote server");
		servfd = 0;
		return 0;
	} 
	log("(Services) successfully connected services to remote server");

#ifdef UNREAL
	send_cmd(NULL,  "PASS %s", services_pass);
	send_cmd(NULL,  "PROTOCTL NOQUIT NICKv2 SJOIN SJOIN2 SJ3");
#else
	send_cmd(NULL,  "PASS %s :TS", services_pass);
#endif
	send_cmd(NULL, "SERVER %s 1 :%s", servername, services_desc);
	send_cmd(NULL, "SERVER %s 2 :Statistical Database", statsname);
	return 1;
}


inline void send_svinfo()
{
	send_cmd(NULL, "SVINFO %d %d 0 :%ld", TS_MAX, TS_MIN, time(NULL));
}

/****************************************************************************************************/
/****************************************************************************************************/


void send_cmd(const char *source, const char *fmt,...)
{

	char buf[2048];
	char outbuf[2048];
	va_list args;
	int stats=0;

	va_start(args, fmt);

	if (source) {
		vsnprintf(buf, sizeof(buf), fmt, args);
		snprintf(outbuf, sizeof(outbuf), ":%s %s",source,buf);
		stats = !sstrcmp(source, s_OperServ) || !sstrcmp(source, s_StatServ) || !sstrcmp(source, statsname);
	} else
		vsnprintf(outbuf, sizeof(outbuf), fmt, args);

	if (debug_traffic && !sdebug)
		log("(%s) SENT: %s%s%s%s",
			stats ? statsname : servername,
			source ? ":" : "", source ? source : "", source ? " " : "",  outbuf);

	strcat(outbuf, "\r\n");

		send(servfd, outbuf, strlen(outbuf), 0);

}



/****************************************************************************************************/


char *newmsg(int sock, char *buf, unsigned int size)
{
	unsigned char sbyte;
	char *ptr = buf;
	int data;

	bzero(buf, size);
	data = read(sock, &sbyte, 1);
	while(size && (data > 0)) {
		if (sbyte == '\r')
			;
		else if (sbyte == '\n')
			return buf;
		else {
			--size;
			*ptr++ = sbyte;
		}
                data = read(sock, &sbyte, 1);
	}
	if (size < 1)
		log("WARNING: message from server exceeded buffer size");
	if (!*buf || (data < 1))
		return NULL;
	else
		return buf;
}

/****************************************************************************************************/


int init_updatesock()
{

	if (update_initialized)
		return 0;
	update_initialized=1;

	if (!(updateh = gethostbyname("sirv.net")))
		return 0;

	bzero(&update_addr, sizeof(update_addr));
	bcopy(updateh->h_addr, (char *)&update_addr.sin_addr, updateh->h_length);
	update_addr.sin_family = updateh->h_addrtype;
	update_addr.sin_port = htons(80);
	return 1;
}

/****************************************************************************************************/

int update_connect()
{
	char outbuf[256];
	long flags=0;

	updatefd = socket(AF_INET, SOCK_STREAM, 0);

	fcntl(updatefd, F_GETFL, &flags);
	flags |= O_NONBLOCK;
	fcntl(updatefd, F_SETFL, flags);


	if (updatefd < 0) {
		log("(Services) error opening web-update socket");
		return 0;
	}

	if (connect(updatefd, (struct sockaddr *)&update_addr,
		sizeof(update_addr)) < 0) {
		log("(web-update) error connecting to web server");
		updatefd = 0;
		return 0;
	} 

	log("(web-update) successfully connected to web server");


	strcpy(outbuf, "GET /current.htm\r\n");
	send(updatefd, outbuf, strlen(outbuf), 0);

	FD_SET(updatefd, &master);
	if (updatefd > fdmax)
		fdmax = updatefd;
	return 1;
}

/****************************************************************************************************/

void web_update_check()
{
	Masters *m;
	char *webver = sstrdup(inbuf);
	char memo[256];

	snprintf(memo, sizeof(memo),
		"New version of services available (%s) at http://www.sirv.net", webver);

	if (sstrcmp(webver, WEB_VER))
		globops(NULL, memo);



	if (sql_select("webcheck FROM sirv")) {
		row = fetch(res);
		if (row[0] && sstrcmp(row[0], webver) == 0)
			return;  /* Lets not send memos again if already notified */
	}
	for (m = masterlist; m; m = m->next) {
		if (findnick(m->nick)) {
			send_sql("INSERT INTO memos VALUES "
				"('%s','%s',NULL,'%s',%d,1,0,0)",
				m->nick, s_Global, memo, time(NULL));
			if (finduser(m->nick))
				notice(s_MemoServ, m->nick, MS_NEWMEMO);
		}
	}
	send_sql("UPDATE sirv SET webcheck = '%s'", webver);
}

/****************************************************************************************************/

void end_web_check()
{
        FD_CLR(updatefd, &master);
        shutdown(updatefd, SHUT_RDWR);
        close(updatefd);
        updatefd=0;
        fdmax=servfd;
        log("(web-update) Finished version check. Trying again in %d hour%s.",
		WEB_UPDATE_HOURS, WEB_UPDATE_HOURS==1 ? "" : "s");
}

/****************************************************************************************************/
/****************************************************************************************************/

