/*
 * SirvNET Services are copyright (c) 1998-2006 Trevor Klingbeil.
 *    http://www.sirv.net  ::  ftp.sirv.net
 * This program is copyrighted software; see the file COPYING for details.
 * Parts of this file are originally based on EsperNet Services(c) by Andy Church.
 */


#include "services.h"
#include "log.h"
#include "timeout.h"


User *userlist[256];

unsigned int usercnt=0, maxusercnt=0, opcnt=0;


UserNode *usernodes[256];


/*********************************************************************************/
/*********************************************************************************/

int check_r_mode(User *user)
{
	struct u_idlist *nl;

	for (nl = user->nicks; nl; nl = nl->next) {
		if (sstrcmp(nl->nick, user->nick) == 0) {
			send_cmd(NULL, "SVSMODE %s +r", user->nick);
			return 1;
		}
	}
	return 0;
}

/*********************************************************************************/

static void find_idnicks(User *user)
{
	int i=0, num_rows=0;

/* We check for nick.stamp and oper.stamp for security. If we didn't check for oper.stamp,
   and services split and came back we'd be identified to oper services even if we're only
   identified to nickserv */

	if (is_oper(user->nick) && sql_select("oper.name,oper.access FROM oper,nick WHERE nick.stamp = '%d' "
		"AND oper.stamp = '%d' AND oper.name=nick.name", user->timestamp, user->timestamp)) {
		row = fetch(res);
		user->os_nick = sstrdup(row[0]);
		user->os_level = atoi(row[1]);
	}

	if ((num_rows = sql_select("name FROM nick WHERE stamp = '%d'", user->timestamp))) {
		for (i=0; i < num_rows; i++) {
			row = fetch(res);
			add_idnick(user, row[0]);
		}
	}
}

/*********************************************************************************/


void newuser_routines(User *u)
{

#ifndef SKELETON
	if (LOGON_NEWS)
		send_logonnews(u);
#endif

}


/*********************************************************************************/


static void new_rootnode(int alpha, unsigned long hvalue, UserNode *newnode)
{
	usernodes[alpha] = scalloc(sizeof(UserNode), 1);
	usernodes[alpha]->value = 50000000;
	usernodes[alpha]->u = scalloc(sizeof(User), 1);
	strncpy(usernodes[alpha]->u->nick, "*", 1);
	if (hvalue < 50000000)
		usernodes[alpha]->left = newnode;
	else
		usernodes[alpha]->right = newnode;
	newnode->back = usernodes[alpha];
}

/*********************************************************************************/

static void retree_usernode(User *user)
{
	UserNode *nptr, *movenode;
	unsigned long hvalue;;


	if (user->node->back) {
		if (user->node->back->right == user->node)
			user->node->back->right = user->node->right;
		else
			user->node->back->left = user->node->right;
		if (user->node->right)
			user->node->right->back = user->node->back;
	}

	if (user->node->left) {
		nptr = user->node->back;
		movenode = user->node->left;

		while (nptr != NULL) {
			if (movenode->value < nptr->value) {
				if (nptr->left)
					nptr = nptr->left;
				else {
					nptr->left = movenode;
					break;
				}
			} else if (movenode->value >= nptr->value) {
				if (nptr->right)
					nptr = nptr->right;
				else {
					nptr->right = movenode;
					break;
				}
			}
		}
		movenode->back = nptr;
	}

	/* retree */

	nptr = usernodes[toupper(*user->nick) & 0x000000FF];
	movenode = user->node;

	nick_hash(user->nick, &hvalue);
	movenode->value = hvalue;
	movenode->right = NULL;
	movenode->left = NULL;

	if (!nptr) {
		new_rootnode(toupper(*user->nick) & 0x000000FF, hvalue, user->node);
		return;
	}

	while (nptr != NULL) {
		if (movenode->value < nptr->value) {
			if (nptr->left)
				nptr = nptr->left;
			else {
				nptr->left = movenode;
				break;
			}
		} else if (movenode->value >= nptr->value) {
			if (nptr->right)
				nptr = nptr->right;
			else {
				nptr->right = movenode;
				break;
			}
		}
	}
	movenode->back = nptr;
}

/*********************************************************************************/

/* If needed, move user structure to appropriate linked list.
   When finished, retree the child usernodes. */

static void move_user(User *user, int oldalpha)
{
	int alpha = toupper(*user->nick) & 0x000000FF;

	if (alpha != oldalpha) {
		if (user->prev)
			user->prev->next = user->next;
		else
			userlist[oldalpha] = user->next;
		if (user->next)
			user->next->prev = user->prev;


		user->prev = NULL;
		user->next = userlist[alpha];
		if (userlist[alpha])
			userlist[alpha]->prev = user;
		userlist[alpha] = user;
	}
	retree_usernode(user);
}

/*********************************************************************************/

static User *new_user(const char *nick)
{
	User *user = scalloc(sizeof(User), 1);
	UserNode *un, *newnode = scalloc(sizeof(UserNode), 1);
	int alpha = toupper(*nick) & 0x000000FF;
	unsigned long hvalue=0;


	if (!nick)
		nick = "";
	strncpy(user->nick, nick, NICKMAX);
	user->floodlev = 0;

	user->next = userlist[alpha];
	if (userlist[alpha])
		userlist[alpha]->prev = user;
	userlist[alpha] = user;

	usercnt++;
	if (usercnt > maxusercnt)
		maxusercnt = usercnt;



	nick_hash((char *)nick, &hvalue);

	newnode->u = user;
	newnode->u->node = newnode;
	newnode->value = hvalue;


	if (!(un = usernodes[alpha])) {
		new_rootnode(alpha, hvalue, newnode);
		return user;
	}


	while (un != NULL) {
		if (hvalue < un->value) {
			if (un->left)
				un = un->left;
			else {
				un->left = newnode;
				break;
			}
		} else if (hvalue >= un->value) {
			if (un->right)
				un = un->right;
			else {
				un->right = newnode;
				break;
			}
		}
	}
	newnode->back = un;
	return user;
}


/*********************************************************************************/

#ifdef BAHAMUT
	#define n_umodes 3
	#define n_username 4
	#define n_hostname 5
	#define n_servername 6
	#define n_sstamp 7
#else /* UNREAL */
	#define n_umodes 7
	#define n_username 3
	#define n_hostname 4
	#define n_servername 5
	#define n_sstamp 6
#endif

void do_nick(char *source, int ac, char **av)
{
	User *user;
	int newuser=0;

	if (ac < 2)
		return;

	if (!*source) {    /* new user, create a structure */
		char *mode[2];


		if (check_akill(av[0], av[4], av[5]))
			return;

		if (debug)
			log("new user: %s", av[0]);

		user = new_user(av[0]);
		user->signon = atol(av[2]);
		user->username = sstrdup(av[n_username]);
		user->host = sstrdup(av[n_hostname]);
		user->server = sstrdup(av[n_servername]);
		user->timestamp = atol(av[n_sstamp]);
#ifdef BAHAMUT
		user->ip = strtoul(av[8], NULL, 10);
#else /* UNREAL */
		user->ip = sstrdup(av[8]); /* virtual host */
#endif
		user->realname = sstrdup(av[9]);
		user->invalid_nicks = 0;
		user->invalid_chans = 0;
		user->invalid_time = time(NULL);
		user->os_level = 0;
		user->lastreg = 0;
		user->lastmemo = 0;
		user->nchanges = 0;
		user->fchans = NULL; user->nicks = NULL; user->chans = NULL;
		user_add_server(user);
		mode[0] = sstrdup(user->nick);
		mode[1] = sstrdup(av[n_umodes]);
		u_mode(user->nick, 2, mode);
		free(mode[0]); free(mode[1]);
		if (!user->timestamp && uburst)
			newuser_routines(user);

		else if (user->timestamp) /* search for nicks already identified for */
			find_idnicks(user);
		newuser=1;


		if (DETECT_CLONES)
			check_clones(user);

	} else {     /* old user changing nicks */
		user = finduser(source);

		if (!user) {
			if (debug)
				log("user: NICK from nonexistent nick %s: %s", source,
					merge_args(ac, av));
			return;
		}
		if (debug)
			log("nick: %s changed nick to %s", source, av[0]);


		cancel_user(user);
		strncpy(user->nick, av[0], NICKMAX);
		user->signon = strtoul(av[ac-1], NULL, 10);
		move_user(user, toupper(*source) & 0x000000FF);
		user->nchanges++;
	}
	user->mode &= ~U_RECOGNIZED;
	user->mode &= ~U_MEMO_NOTIFIED;

#ifndef SKELETON
	if (uburst) {
		validate_user(user,newuser);
		check_memos(user, 0);
	}
#endif

}

/*********************************************************************************/

void delete_user(User *user)
{
	UserNode *un = user->node;
	struct u_idlist *nl, *nl2;
	struct u_founderlist *fl, *fl2;
	int alpha = toupper(*user->nick) & 0x000000FF;


#ifdef BAHAMUT
	userhost_list(user, DEL, 0, 0);
#endif

	if (un->back) {
		if (un->back->right == un)
			un->back->right = un->right;
		else
			un->back->left = un->right;
		if (un->right)
			un->right->back = un->back;
	}

	if (un->left) {    /* retree */
		UserNode *nptr = un->back, *movenode = un->left;

		while (nptr != NULL) {
			if (movenode->value < nptr->value) {
				if (nptr->left)
					nptr = nptr->left;
				else {
					nptr->left = movenode;
					break;
				}
			} else if (movenode->value >= nptr->value) {
				if (nptr->right)
					nptr = nptr->right;
				else {
					nptr->right = movenode;
					break;
				}
			}
		}
		movenode->back = nptr;
	}
	free(un);


	if (nick_registered(user->nick) && is_recognized(user->nick, NULL))
		send_sql("UPDATE nick SET lasttime = %d WHERE name = '%s'", 
			time(NULL), user->nick);


	if (user->mode & UMODE_O)
		opcnt--;

	if (user->chans)
		delete_uchans(user);

	nl = user->nicks;
	while (nl) {
		nl2 = nl->next;
		send_sql("DELETE FROM memos WHERE recipient='%s' AND del=1", nl->nick);
		free(nl->nick); free(nl);
		nl = nl2;
	}
	fl = user->fchans;
	while (fl) {
		fl2 = fl->next;
		free(fl);
		fl = fl2;
	}



	free(user->username);
	free(user->host);
	free(user->realname);
	free(user->server);
	if (user->os_nick)
		free(user->os_nick);


	if (user->prev)
		user->prev->next = user->next;
	else
		userlist[alpha] = user->next;
	if (user->next)
		user->next->prev = user->prev;
	free(user);

	usercnt--;
}

/*********************************************************************************/

void do_quit(char *source, int ac, char **av)
{
	User *user;


	user = finduser(source);
	if (!user || !source)
		return;

	if (debug)
		log("QUIT: %s", source);


	user_del_server(user);
	delete_user(user);
}

/*********************************************************************************/

void do_kill(char *source, int ac, char **av)
{
	User *user;

	if (check_servii_kill(source, ac, av))
		return;

	user = finduser(av[0]);
	if (!user)
		return;

	if (debug)
		log("KILL: %s -> %s", source, av[0]);

	user_del_server(user);
	delete_user(user);


}

/*********************************************************************************/

int is_oper(const char *nick)
{
	User *user = finduser(nick);
	return user && (user->mode & UMODE_O);
}

/*********************************************************************************/

User *finduser(const char *nick)
{
	UserNode *un;
	unsigned long hvalue=0;
	unsigned int alpha = toupper(*nick) & 0x000000FF;


	if ((alpha > 256) || (alpha < 0) || !(un = usernodes[alpha]))
		return NULL;

	nick_hash((char *)nick, &hvalue);


	while (un != NULL) {
		if ((hvalue == un->value) && sstrcmp(nick, un->u->nick) == 0)
			return un->u;
		un = (hvalue<un->value) ? un->left : un->right;
        }
        return NULL;

}

/*********************************************************************************/


int match_usermask(const char *mask, User *user)
{
	char *mask2 = sstrdup(mask);
	char *nick=0, *username, *host, *nick2=0, *host2;
	int result;

	if (strchr(mask2, '!')) {
		nick = strlower(strtok(mask2, "!"));
		username = strtok(NULL, "@");
	} else {
		nick = NULL;
		username = strtok(mask2, "@");
	}
	host = strtok(NULL, "");
	if (!username || !host) {
		free(mask2);
		return 0;
	}
	strlower(host);
	if (nick)
		nick2 = strlower(sstrdup(user->nick));
	host2 = strlower(sstrdup(user->host));
	result = (nick ? match_wild(nick, nick2) : 1) &&
	match_wild(username, user->username) &&
	match_wild(host, host2);
	free(mask2);
	if (nick)
		free(nick2); 
	free(host2);

	return result;
}


/*********************************************************************************/

char *create_mask(User *u)
{
	char *mask, *s, *end;

	end = mask = smalloc(strlen(u->username) + strlen(u->host) + 2);
	end += sprintf(end, "%s@", u->username);
	if (strspn(u->host, "0123456789.") == strlen(u->host)
		&& (s = strchr(u->host, '.'))
		&& (s = strchr(s+1, '.'))
		&& (s = strchr(s+1, '.'))
		&& (   !strchr(s+1, '.'))) {    /* IP addr */
		s = sstrdup(u->host);
		*strrchr(s, '.') = 0;
		sprintf(end, "%s.*", s);
		free(s);
	} else {
		if ((s = strchr(u->host, '.')) && strchr(s+1, '.')) {
			s = sstrdup(strchr(u->host, '.')-1);
			*s = '*';
		} else {
			s = sstrdup(u->host);
		}
		strcpy(end, s);
		free(s);
	}
	return mask;
}


/*********************************************************************************/
/*********************************************************************************/

